"""Vid Club Cloud API connector."""

from datetime import datetime, timedelta
from typing import Any, Literal

import pandas as pd

from viadot.exceptions import ValidationError
from viadot.sources.base import Source
from viadot.utils import handle_api_response


class VidClub(Source):
    """A class implementing the Vid Club API.

    Documentation for this API is located at: https://evps01.envoo.net/vipapi/
    There are 4 endpoints where to get the data.
    """

    def __init__(
        self,
        *args,
        endpoint: Literal["jobs", "product", "company", "survey"] | None = None,
        from_date: str = "2022-03-22",
        to_date: str | None = None,
        items_per_page: int = 100,
        region: Literal["bg", "hu", "hr", "pl", "ro", "si", "all"] | None = None,
        days_interval: int = 30,
        cols_to_drop: list[str] | None = None,
        vid_club_credentials: dict[str, Any] | None = None,
        validate_df_dict: dict | None = None,
        timeout: int = 3600,
        **kwargs,
    ):
        """Create an instance of VidClub.

        Args:
            endpoint (Literal["jobs", "product", "company", "survey"], optional): The
                endpoint source to be accessed. Defaults to None.
            from_date (str, optional): Start date for the query, by default is the
                oldest date in the data 2022-03-22.
            to_date (str, optional): End date for the query. By default None,
                which will be executed as datetime.today().strftime("%Y-%m-%d") in code.
            items_per_page (int, optional): Number of entries per page. Defaults to 100.
            region (Literal["bg", "hu", "hr", "pl", "ro", "si", "all"], optional):
                Region filter for the query. Defaults to None
                (parameter is not used in url). [December 2023 status: value 'all'
                does not work for company and jobs]
            days_interval (int, optional): Days specified in date range per API call
                (test showed that 30-40 is optimal for performance). Defaults to 30.
            cols_to_drop (List[str], optional): List of columns to drop.
                Defaults to None.
            vid_club_credentials (Dict[str, Any], optional): Stores the credentials
                information. Defaults to None.
            validate_df_dict (dict, optional): A dictionary with optional list of tests
                to verify the output
                dataframe. If defined, triggers the `validate_df` task from task_utils.
                Defaults to None.
            timeout (int, optional): The time (in seconds) to wait while running this
                task before a timeout occurs. Defaults to 3600.
        """
        self.endpoint = endpoint
        self.from_date = from_date
        self.to_date = to_date
        self.items_per_page = items_per_page
        self.region = region
        self.days_interval = days_interval
        self.cols_to_drop = cols_to_drop
        self.vid_club_credentials = vid_club_credentials
        self.validate_df_dict = validate_df_dict
        self.timeout = timeout

        self.headers = {
            "Authorization": "Bearer " + vid_club_credentials["token"],
            "Content-Type": "application/json",
        }

        super().__init__(credentials=vid_club_credentials, *args, **kwargs)  # noqa: B026

    def build_query(
        self,
        from_date: str,
        to_date: str,
        api_url: str,
        items_per_page: int,
        endpoint: Literal["jobs", "product", "company", "survey"] | None = None,
        region: Literal["bg", "hu", "hr", "pl", "ro", "si", "all"] | None = None,
    ) -> str:
        """Builds the query from the inputs.

        Args:
            from_date (str): Start date for the query.
            to_date (str): End date for the query, if empty, will be executed as
                datetime.today().strftime("%Y-%m-%d").
            api_url (str): Generic part of the URL to Vid Club API.
            items_per_page (int): number of entries per page.
            endpoint (Literal["jobs", "product", "company", "survey"], optional):
                The endpoint source to be accessed. Defaults to None.
            region (Literal["bg", "hu", "hr", "pl", "ro", "si", "all"], optional):
                Region filter for the query. Defaults to None
                (parameter is not used in url). [December 2023 status: value 'all'
                does not work for company and jobs]

        Returns:
            str: Final query with all filters added.

        Raises:
            ValidationError: If any source different than the ones in the list are used.
        """
        if endpoint in ["jobs", "product", "company"]:
            region_url_string = f"&region={region}" if region else ""
            url = (
                f"""{api_url}{endpoint}?from={from_date}&to={to_date}"""
                f"""{region_url_string}&limit={items_per_page}"""
            )
        elif endpoint == "survey":
            url = f"{api_url}{endpoint}?language=en&type=question"
        else:
            msg = "Pick one these sources: jobs, product, company, survey"
            raise ValidationError(msg)
        return url

    def intervals(
        self, from_date: str, to_date: str, days_interval: int
    ) -> tuple[list[str], list[str]]:
        """Breaks dates range into smaller by provided days interval.

        Args:
            from_date (str): Start date for the query in "%Y-%m-%d" format.
            to_date (str): End date for the query, if empty, will be executed as
                datetime.today().strftime("%Y-%m-%d").
            days_interval (int): Days specified in date range per api call
                (test showed that 30-40 is optimal for performance).

        Returns:
            List[str], List[str]: Starts and Ends lists that contains information
                about date ranges for specific period and time interval.

        Raises:
            ValidationError: If the final date of the query is before the start date.
        """
        if to_date is None:
            to_date = datetime.today().strftime("%Y-%m-%d")

        end_date = datetime.strptime(to_date, "%Y-%m-%d").date()
        start_date = datetime.strptime(from_date, "%Y-%m-%d").date()

        from_date_obj = datetime.strptime(from_date, "%Y-%m-%d")

        to_date_obj = datetime.strptime(to_date, "%Y-%m-%d")
        delta = to_date_obj - from_date_obj

        if delta.days < 0:
            msg = "to_date cannot be earlier than from_date."
            raise ValidationError(msg)

        interval = timedelta(days=days_interval)
        starts = []
        ends = []

        period_start = start_date
        while period_start < end_date:
            period_end = min(period_start + interval, end_date)
            starts.append(period_start.strftime("%Y-%m-%d"))
            ends.append(period_end.strftime("%Y-%m-%d"))
            period_start = period_end
        if len(starts) == 0 and len(ends) == 0:
            starts.append(from_date)
            ends.append(to_date)
        return starts, ends

    def check_connection(
        self,
        endpoint: Literal["jobs", "product", "company", "survey"] | None = None,
        from_date: str = "2022-03-22",
        to_date: str | None = None,
        items_per_page: int = 100,
        region: Literal["bg", "hu", "hr", "pl", "ro", "si", "all"] | None = None,
        url: str | None = None,
    ) -> tuple[dict[str, Any], str]:
        """Initiate first connection to API to retrieve piece of data.

        With information about type of pagination in API URL.
        This option is added because type of pagination for endpoints is being changed
        in the future from page number to 'next' id.

        Args:
            endpoint (Literal["jobs", "product", "company", "survey"], optional):
                The endpoint source to be accessed. Defaults to None.
            from_date (str, optional): Start date for the query, by default is the
                oldest date in the data 2022-03-22.
            to_date (str, optional): End date for the query. By default None,
                which will be executed as datetime.today().strftime("%Y-%m-%d") in code.
            items_per_page (int, optional): Number of entries per page.
                100 entries by default.
            region (Literal["bg", "hu", "hr", "pl", "ro", "si", "all"], optional):
                Region filter for the query. Defaults to None
                (parameter is not used in url). [December 2023 status: value 'all'
                does not work for company and jobs]
            url (str, optional): Generic part of the URL to Vid Club API.
                Defaults to None.

        Returns:
            Tuple[Dict[str, Any], str]: Dictionary with first response from API with
                JSON containing data and used URL string.

        Raises:
            ValidationError: If from_date is earlier than 2022-03-22.
            ValidationError: If to_date is earlier than from_date.
        """
        if from_date < "2022-03-22":
            msg = "from_date cannot be earlier than 2022-03-22."
            raise ValidationError(msg)

        if to_date < from_date:
            msg = "to_date cannot be earlier than from_date."
            raise ValidationError(msg)

        if url is None:
            url = self.credentials["url"]

        first_url = self.build_query(
            endpoint=endpoint,
            from_date=from_date,
            to_date=to_date,
            api_url=url,
            items_per_page=items_per_page,
            region=region,
        )
        headers = self.headers
        response = handle_api_response(url=first_url, headers=headers, method="GET")
        response = response.json()
        return (response, first_url)

    def get_response(
        self,
        endpoint: Literal["jobs", "product", "company", "survey"] | None = None,
        from_date: str = "2022-03-22",
        to_date: str | None = None,
        items_per_page: int = 100,
        region: Literal["bg", "hu", "hr", "pl", "ro", "si", "all"] | None = None,
    ) -> pd.DataFrame:
        """Basing on the pagination type retrieved using check_connection function.

        It gets the response from the API queried and transforms it into DataFrame.

        Args:
            endpoint (Literal["jobs", "product", "company", "survey"], optional):
                The endpoint source to be accessed. Defaults to None.
            from_date (str, optional): Start date for the query, by default is the
                oldest date in the data 2022-03-22.
            to_date (str, optional): End date for the query. By default None,
                which will be executed as datetime.today().strftime("%Y-%m-%d") in code.
            items_per_page (int, optional): Number of entries per page.
                100 entries by default.
            region (Literal["bg", "hu", "hr", "pl", "ro", "si", "all"], optional):
                Region filter for the query. Defaults to None
                (parameter is not used in url). [December 2023 status: value 'all'
                does not work for company and jobs]

        Returns:
            pd.DataFrame: Table of the data carried in the response.

        Raises:
            ValidationError: If any source different than the ones in the list are used.
        """
        headers = self.headers
        if endpoint not in ["jobs", "product", "company", "survey"]:
            msg = "The source has to be: jobs, product, company or survey"
            raise ValidationError(msg)
        if to_date is None:
            to_date = datetime.today().strftime("%Y-%m-%d")

        response, first_url = self.check_connection(
            endpoint=endpoint,
            from_date=from_date,
            to_date=to_date,
            items_per_page=items_per_page,
            region=region,
        )

        if isinstance(response, dict):
            keys_list = list(response.keys())
        elif isinstance(response, list):
            keys_list = list(response[0].keys())
        else:
            keys_list = []

        ind = "next" in keys_list

        if "data" in keys_list:
            df = pd.json_normalize(response["data"])
            df = pd.DataFrame(df)
            length = df.shape[0]
            page = 1

            while length == items_per_page:
                if ind is True:
                    next_page = response["next"]
                    url = f"{first_url}&next={next_page}"
                else:
                    page += 1
                    url = f"{first_url}&page={page}"
                response_api = handle_api_response(
                    url=url, headers=headers, method="GET"
                )
                response = response_api.json()
                df_page = pd.json_normalize(response["data"])
                df_page = pd.DataFrame(df_page)
                if endpoint == "product":
                    df_page = df_page.transpose()
                length = df_page.shape[0]
                df = pd.concat((df, df_page), axis=0)
        else:
            df = pd.DataFrame(response)

        return df

    def to_df(
        self,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
    ) -> pd.DataFrame:
        """Looping get_response and iterating by date ranges defined in intervals.

        Stores outputs as DataFrames in a list. At the end, daframes are concatenated
        in one and dropped duplicates that would appear when quering.

        Args:
            if_empty (Literal["warn", "skip", "fail"], optional): What to do if a fetch
                produce no data. Defaults to "warn

        Returns:
            pd.DataFrame: Dataframe of the concatanated data carried in the responses.
        """
        starts, ends = self.intervals(
            from_date=self.from_date,
            to_date=self.to_date,
            days_interval=self.days_interval,
        )

        dfs_list = []
        if len(starts) > 0 and len(ends) > 0:
            for start, end in zip(starts, ends, strict=False):
                self.logger.info(f"ingesting data for dates [{start}]-[{end}]...")
                df = self.get_response(
                    endpoint=self.endpoint,
                    from_date=start,
                    to_date=end,
                    items_per_page=self.items_per_page,
                    region=self.region,
                )
                dfs_list.append(df)
                if len(dfs_list) > 1:
                    df = pd.concat(dfs_list, axis=0, ignore_index=True)
                else:
                    df = pd.DataFrame(dfs_list[0])
        else:
            df = self.get_response(
                endpoint=self.endpoint,
                from_date=self.from_date,
                to_date=self.to_date,
                items_per_page=self.items_per_page,
                region=self.region,
            )
        list_columns = df.columns[df.map(lambda x: isinstance(x, list)).any()].tolist()
        for i in list_columns:
            df[i] = df[i].apply(lambda x: tuple(x) if isinstance(x, list) else x)
        df.drop_duplicates(inplace=True)

        if self.cols_to_drop is not None:
            if isinstance(self.cols_to_drop, list):
                try:
                    self.logger.info(
                        f"Dropping following columns: {self.cols_to_drop}..."
                    )
                    df.drop(columns=self.cols_to_drop, inplace=True, errors="raise")
                except KeyError:
                    self.logger.exception(
                        f"""Column(s): {self.cols_to_drop} don't exist in the DataFrame.
                        No columns were dropped. Returning full DataFrame..."""
                    )
                    self.logger.info(f"Existing columns: {df.columns}")
            else:
                msg = "Provide columns to drop in a List."
                raise TypeError(msg)

        if df.empty:
            self._handle_if_empty(if_empty=if_empty)

        return df



"""Task for downloading data from Matomo API to a pandas DataFrame."""

from typing import Any, Literal

import pandas as pd
from prefect import task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources.matomo import Matomo


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=60 * 60)
def matomo_to_df(
    url: str,
    top_level_fields: list[str],
    record_path: str | list[str],
    params: dict[str, Any],
    config_key: str | None = None,
    credentials_secret: str | None = None,
    record_prefix: str | None = None,
    if_empty: Literal["warn", "skip", "fail"] = "warn",
    tests: dict[str, Any] | None = None,
) -> pd.DataFrame:
    """Task to download data from Matomo API to a pandas DataFrame.

    Args:
        url (str): The base URL of the Matomo instance.
        top_level_fields (list[str]): List of top level fields to get from the API JSON.
        record_path (str | list[str]): The path field to the records in the API
            response.Could be handled as a list of path + fields to extract:
                    record_path = 'actionDetails'
                    record_path = ['actionDetails', 'eventAction']
        config_key (Optional[str], optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        credentials_secret (str, optional): The name of the secret that stores Matomo
            credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        params (dict[str, Any]): Parameters for the API request.
                Necessary params and their examples are:
                    "module": "API",
                    "method": "Live.getLastVisitsDetails",
                    "idSite": "53",
                    "period": "range",
                    "date": ("<<yesterday>>", "<<yesterday>>"),
                    "format": "JSON",
        record_prefix (Optional[str], optional): A prefix for the record path fields.
            For example: "action_". Defaults to None.
        if_empty (Literal["warn", "skip", "fail"], optional): What to do if the
            query returns no data. Defaults to "warn".
        tests (dict[str, Any], optional): A dictionary with optional list of tests
            to verify the output dataframe. If defined, triggers the `validate`
            function from viadot.utils. Defaults to None.

    Examples:
        data_frame = matomo_to_df(
            url="https://matomo.example.com",
            top_level_fields=["idSite", "visitorId", "visitIp"],
            record_path="actionDetails",
            record_prefix="action_",
            config_key="matomo_prod",
            params={
                "module": "API",
                "method": "Live.getLastVisitsDetails",
                "idSite": "53",
                "period": "range",
                "date": ("<<yesterday>>", "<<today>>"),
                "format": "JSON"
            },
            record_prefix="action_",
            if_empty="warn"
        )

    Raises:
        MissingSourceCredentialsError: If no credentials have been provided.
        ValueError: If api_token is not found in credentials or if no data
            has been fetched.

    Returns:
        pd.DataFrame: The Matomo data as a pandas DataFrame.
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    credentials = get_credentials(credentials_secret)
    matomo = Matomo(
        credentials=credentials,
        config_key=config_key,
    )

    params["date"] = matomo.format_date_range(params["date"])

    # Fetch the data using credentials
    data = matomo.fetch_data(
        api_token=matomo.credentials["api_token"],
        url=url,
        params=params,
    )

    # Convert to DataFrame with the specified parameters
    return matomo.to_df(
        data=data,
        top_level_fields=top_level_fields,
        record_path=record_path,
        record_prefix=record_prefix,
        if_empty=if_empty,
        tests=tests,
    )



"""Build specified dbt model(s)."""

import os
import shutil

from prefect import flow, task

from viadot.orchestration.prefect.tasks import clone_repo, dbt_task
from viadot.orchestration.prefect.utils import get_credentials, with_flow_timeout_param


@task(cache_policy=None)
def _cleanup_repo(dbt_repo_dir_name: str) -> None:
    """Remove a repo folder.

    Args:
        dbt_repo_dir_name (str): The name of the temporary folder.
    """
    shutil.rmtree(dbt_repo_dir_name, ignore_errors=True)  # Delete folder on run


@flow(
    name="Transform",
    description="Build specified dbt model(s).",
)
@with_flow_timeout_param()
def transform(
    dbt_project_path: str,
    dbt_repo_url: str | None = None,
    dbt_repo_url_secret: str | None = None,
    dbt_repo_branch: str | None = None,
    token_secret: str | None = None,
    local_dbt_repo_path: str | None = None,
    dbt_selects: dict[str, str] | None = None,
    dbt_target: str | None = None,
) -> None:
    """Build specified dbt model(s).

    This flow implements a simplified version of the `transform_and_catalog()` flow,
    excluding metadata handling, source freshness checks, and stateful operations.

    Args:
        dbt_project_path (str): The path to the dbt project (the directory containing
            the `dbt_project.yml` file).
        dbt_repo_url (str, optional): The URL for cloning the dbt repo with relevant
            dbt project.
        dbt_repo_url_secret (str, optional): Alternatively to above, the secret
            containing `dbt_repo_url`.
        dbt_repo_branch (str, optional): The branch of the dbt repo to use.
        token_secret (str, optional): The name of the secret storing the git token.
            Defaults to None.
        local_dbt_repo_path (str, optional): The path where to clone the repo to.
        dbt_selects (dict, optional): Valid
            [dbt node selection](https://docs.getdbt.com/reference/node-selection/syntax)
            expressions. Valid keys are `run`, `test`, and `source_freshness`. The test
                select expression is taken from run's, as long as run select is
                provided. Defaults to None.
        dbt_target (str): The dbt target to use. If not specified, the default dbt
            target (as specified in `profiles.yaml`) will be used. Defaults to None.

    Returns:
        list[str]: Lines from stdout of the `upload_metadata` task as a list.

    Examples:
        # Build a single model
        ```python
        import os
        from prefect_viadot.flows import transform_and_catalog

        my_dbt_project_path = os.path.expanduser("~/dbt/my_dbt_project")

        transform_and_catalog(
            dbt_project_path=my_dbt_project_path,
            dbt_selects={"run": "my_model"}
        )
        ```

        Some common `dbt_select` patterns:
        - build a model and all its downstream dependencies: `dbt_select="my_model+"`
        - build all models in a directory: `dbt_select="models/my_project"`
        ```
    """
    dbt_repo_url = dbt_repo_url or get_credentials(dbt_repo_url_secret)
    local_dbt_repo_path = (
        os.path.expandvars(local_dbt_repo_path)
        if local_dbt_repo_path is not None
        else "tmp_dbt_repo_dir"
    )

    clone_repo(
        url=dbt_repo_url,
        checkout_branch=dbt_repo_branch,
        token_secret=token_secret,
        path=local_dbt_repo_path,
    )

    # dbt CLI does not handle passing --target=None
    dbt_target_option = f"-t {dbt_target}" if dbt_target is not None else ""

    # Clean up artifacts from previous runs (`target/` dir and packages)
    dbt_clean_task = dbt_task.with_options(name="dbt_task_clean")
    dbt_clean_up = dbt_clean_task.submit(
        project_path=dbt_project_path,
        command="clean",
    )
    dbt_clean_up.result()
    dbt_pull_deps_task = dbt_task.with_options(name="dbt_task_deps")
    pull_dbt_deps = dbt_pull_deps_task.submit(
        project_path=dbt_project_path,
        command="deps",
    )
    pull_dbt_deps.result()

    run_select = dbt_selects.get("run") if dbt_selects else None
    run_select_safe = f"-s {run_select}" if run_select is not None else ""

    dbt_run_task_func = dbt_task.with_options(name="dbt_task_run")
    run = dbt_run_task_func.submit(
        project_path=dbt_project_path,
        command=f"run {run_select_safe} {dbt_target_option}",
    )
    run.result()

    test_select = dbt_selects.get("test", run_select) if dbt_selects else None
    test_select_safe = f"-s {test_select}" if test_select is not None else ""

    dbt_test_task_func = dbt_task.with_options(name="dbt_task_test")
    test = dbt_test_task_func.submit(
        project_path=dbt_project_path,
        command=f"test {test_select_safe} {dbt_target_option}",
    )
    test.result()

    _cleanup_repo(
        dbt_repo_dir_name=local_dbt_repo_path,
    )



"""'test_bigquery.py'."""

from google.oauth2 import service_account
import numpy as np
import pandas as pd
import pytest

from viadot.exceptions import APIError, CredentialError
from viadot.sources import BigQuery
from viadot.sources.bigquery import BigQueryCredentials


variables = {
    "credentials": {
        "type": "service_account",
        "project_id": "test_project",
        "private_key_id": "some-key-id",
        "private_key": "----------",
        "client_email": "email@test.com",
        "client_id": "client_id",
        "client_x509_cert_url": "https://www.googleapis.com/robot/v1/"
        + "metadata/x509/email@test.com",
    }
}


class TestBigQueryCredentials:
    """Test BigQuery Credentials Class."""

    @pytest.mark.basic
    def test_bigquery_credentials(self):
        """Test BigQuery credentials."""
        BigQueryCredentials(
            type="test_type",
            project_id="",
            private_key_id="test_private_key_id",
            private_key="test_private_key",  # pragma: allowlist secret
            client_email="test_client_email",
            client_id="test_client_id",
            auth_uri="",
            token_uri="",
            auth_provider_x509_cert_url="",
            client_x509_cert_url="",
        )


@pytest.fixture
def mock_service_account_credentials():
    """Mock BigQuery account credentials."""

    class MockCredentials:
        """Mock class."""

        def __init__(self):
            self.project_id = "mock_project_id"

    return MockCredentials()


@pytest.mark.basic
def test_missing_credentials(mocker):
    """Test raise error without BigQuery credentials."""
    mocker.patch("viadot.sources.bigquery.get_source_credentials", return_value=None)
    with pytest.raises(CredentialError):
        BigQuery(config_key="invalid_key")


@pytest.mark.functions
def test_get_list_datasets_query(monkeypatch, mock_service_account_credentials):
    """Test BigQuery `_get_list_datasets_query` method."""
    monkeypatch.setattr(
        service_account.Credentials,
        "from_service_account_info",
        lambda info: mock_service_account_credentials,  # noqa: ARG005
    )
    bigquery = BigQuery(credentials=variables["credentials"])
    expected_query = """SELECT schema_name
                FROM test_project.INFORMATION_SCHEMA.SCHEMATA"""
    query = bigquery._get_list_datasets_query()

    assert query.strip() == expected_query


@pytest.mark.functions
def test_get_list_tables_query(monkeypatch, mock_service_account_credentials):
    """Test BigQuery `_get_list_tables_query` method."""
    monkeypatch.setattr(
        service_account.Credentials,
        "from_service_account_info",
        lambda info: mock_service_account_credentials,  # noqa: ARG005
    )
    bigquery = BigQuery(credentials=variables["credentials"])
    expected_query = """SELECT table_name
                FROM test_project.test_dataset.INFORMATION_SCHEMA.TABLES"""
    query = bigquery._get_list_tables_query(dataset_name="test_dataset")

    assert query.strip() == expected_query


@pytest.mark.functions
def test_list_columns(mocker, monkeypatch, mock_service_account_credentials):
    """Test BigQuery `_list_columns` method."""
    monkeypatch.setattr(
        service_account.Credentials,
        "from_service_account_info",
        lambda info: mock_service_account_credentials,  # noqa: ARG005
    )
    bigquery = BigQuery(credentials=variables["credentials"])

    mock_df = pd.DataFrame({"column_name": ["col1", "col2", "col3"]})
    mocker.patch("pandas_gbq.read_gbq", return_value=mock_df)
    result = bigquery._list_columns(
        dataset_name="test_dataset",
        table_name="test_table",
    )
    expected_result = np.array(["col1", "col2", "col3"])

    np.testing.assert_array_equal(result, expected_result)


@pytest.mark.functions
def test_get_google_bigquery_data_success(
    mocker, monkeypatch, mock_service_account_credentials
):
    """Test BigQuery `_get_google_bigquery_data` method."""
    monkeypatch.setattr(
        service_account.Credentials,
        "from_service_account_info",
        lambda info: mock_service_account_credentials,  # noqa: ARG005
    )
    bigquery = BigQuery(credentials=variables["credentials"])

    mock_df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
    mocker.patch("pandas_gbq.read_gbq", return_value=mock_df)
    result = bigquery._get_google_bigquery_data("SELECT * FROM test_table")

    pd.testing.assert_frame_equal(result, mock_df)


@pytest.mark.functions
def test_get_google_bigquery_data_failure(
    mocker, monkeypatch, mock_service_account_credentials
):
    """Test BigQuery `_get_google_bigquery_data` method failure."""
    monkeypatch.setattr(
        service_account.Credentials,
        "from_service_account_info",
        lambda info: mock_service_account_credentials,  # noqa: ARG005
    )
    bigquery = BigQuery(credentials=variables["credentials"])

    mocker.patch("pandas_gbq.read_gbq", side_effect=APIError("Error"))
    with pytest.raises(APIError):
        bigquery._get_google_bigquery_data("SELECT * FROM test_table")


@pytest.mark.connect
def test_to_df_list_tables(mocker, monkeypatch, mock_service_account_credentials):
    """Test BigQuery `to_df` method with predefined query."""
    monkeypatch.setattr(
        service_account.Credentials,
        "from_service_account_info",
        lambda info: mock_service_account_credentials,  # noqa: ARG005
    )
    bigquery = BigQuery(credentials=variables["credentials"])

    mock_df = pd.DataFrame({"table_name": ["table1", "table2"]})
    mocker.patch.object(BigQuery, "_get_google_bigquery_data", return_value=mock_df)
    df = bigquery.to_df(query="tables", dataset_name="test_dataset")

    pd.testing.assert_frame_equal(df, mock_df)


@pytest.mark.connect
def test_to_df_with_columns_and_date(
    mocker, monkeypatch, mock_service_account_credentials
):
    """Test BigQuery `to_df` method."""
    monkeypatch.setattr(
        service_account.Credentials,
        "from_service_account_info",
        lambda info: mock_service_account_credentials,  # noqa: ARG005
    )
    bigquery = BigQuery(credentials=variables["credentials"])

    def mock_list_columns(self, dataset_name: str, table_name: str):  # noqa: ARG001
        # pylint: disable=unused-argument
        return ["column1", "column2", "test_date_column"]

    monkeypatch.setattr(BigQuery, "_list_columns", mock_list_columns)

    mock_df = pd.DataFrame(
        {"col1": [1, 2], "col2": [3, 4], "date_col": ["2023-01-01", "2023-01-02"]}
    )
    mocker.patch.object(BigQuery, "_get_google_bigquery_data", return_value=mock_df)
    df = bigquery.to_df(
        dataset_name="test_dataset",
        table_name="test_table",
        columns=["col1", "col2"],
        date_column_name="date_col",
        start_date="2023-01-01",
        end_date="2023-01-31",
    )

    pd.testing.assert_frame_equal(df, mock_df)



"""A connector for Cloud For Customers API."""

from copy import deepcopy
import re
from typing import Any, Literal
from urllib.parse import urljoin

import pandas as pd
from pydantic import BaseModel, SecretStr, root_validator
import requests

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.sources.base import Source
from viadot.utils import handle_api_response, validate


class CloudForCustomersCredentials(BaseModel):
    """Cloud for Customers connector credentials validator.

    Validate the credentials.

    Methods:
        is_configured: main method to validate.
    """

    username: str  # eg. username@{tenant_name}.com
    password: SecretStr
    url: str | None = None  # The URL to extract records from.
    report_url: str | None = None  # The URL of a prepared report.

    @classmethod
    @root_validator(pre=True)
    def is_configured(cls, credentials: dict) -> dict:
        """Validate Credentials.

        Args:
            credentials (dict): dictionary with user and password.

        Returns:
            credentials (dict): dictionary with user and password.
        """
        username = credentials.get("username")
        password = credentials.get("password")

        if not (username and password):
            msg = "`username` and `password` credentials are required."
            raise CredentialError(msg)

        return credentials


class CloudForCustomers(Source):
    """Cloud for Customers connector to fetch OData source.

    Args:
        url (str, optional): The URL to the C4C API. For example,
            'https://myNNNNNN.crm.ondemand.com/c4c/v1/'.
        endpoint (str, optional): The API endpoint.
        report_url (str, optional): The URL of a prepared report.
        filter_params (Dict[str, Any], optional): Filtering parameters passed to the
            request. E.g {"$filter": "AccountID eq '1234'"}. More info on:
            https://userapps.support.sap.com/sap/support/knowledge/en/2330688
        credentials (CloudForCustomersCredentials, optional): Cloud for Customers
            credentials.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials.
    """

    DEFAULT_PARAMS = {"$format": "json"}  # noqa: RUF012

    def __init__(
        self,
        *args,
        url: str | None = None,
        endpoint: str | None = None,
        report_url: str | None = None,
        filter_params: dict[str, Any] | None = None,
        credentials: CloudForCustomersCredentials | None = None,
        config_key: str | None = None,
        **kwargs,
    ):
        """Initialize the class with the provided parameters.

        Args:
            *args: Variable length argument list.
            url (str, optional): The base URL for the service.
            endpoint (str, optional): The specific endpoint for the service.
            report_url (str, optional): The URL for the report.
            filter_params (Dict[str, Any], optional): Parameters to filter the report
                data.
            credentials (CloudForCustomersCredentials, optional): Credentials required
                for authentication.
            config_key (Optional[str], optional): A key to retrieve specific
                configuration settings.
            **kwargs: Arbitrary keyword arguments.
        """
        # Credentials logic
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = dict(
            CloudForCustomersCredentials(**raw_creds)
        )  # validate the credentials
        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.url = url or self.credentials.get("url")
        self.report_url = report_url or self.credentials.get("report_url")

        self.is_report = bool(self.report_url)
        self.endpoint = endpoint

        if self.url:
            self.full_url = urljoin(self.url, self.endpoint)
        if filter_params:
            filter_params_merged = self.DEFAULT_PARAMS.copy()
            filter_params_merged.update(filter_params)

            self.filter_params = filter_params_merged
        else:
            self.filter_params = self.DEFAULT_PARAMS

    @staticmethod
    def create_metadata_url(url: str) -> str:
        """Create URL to fetch metadata from.

        Args:
            url (str): The URL to transform to metadata URL.

        Returns:
            meta_url (str): The URL to fetch metadata from.
        """
        start = url.split(".svc")[0]
        url_raw = url.split("?")[0]
        end = url_raw.split("/")[-1]
        return start + ".svc/$metadata?entityset=" + end

    def _extract_records_from_report_url(self, report_url: str) -> list[dict[str, Any]]:
        """Fetch report_url to extract records.

        Args:
            report_url (str): The url to extract records from.

        Returns:
            records (List[Dict[str, Any]]): The records extracted from report_url.
        """
        records = []
        while report_url:
            response = self.get_response(report_url, filter_params=self.filter_params)
            response_json = response.json()
            new_records = self.get_entities(response_json, report_url)
            records.extend(new_records)

            report_url = response_json["d"].get("__next")

        return records

    def _extract_records_from_url(self, url: str) -> list[dict[str, Any]]:
        """Fetch URL to extract records.

        Args:
            url (str): The URL to extract records from.

        Returns:
            records (List[Dict[str, Any]]): The records extracted from URL.
        """
        tmp_full_url = deepcopy(url)
        tmp_filter_params = deepcopy(self.filter_params)
        records = []
        while url:
            response = self.get_response(tmp_full_url, filter_params=tmp_filter_params)
            response_json = response.json()
            if isinstance(response_json["d"], dict):
                # ODATA v2+ API
                new_records = response_json["d"].get("results")
                url = response_json["d"].get("__next", None)
            else:
                # ODATA v1
                new_records = response_json["d"]
                url = response_json.get("__next", None)

            # prevents concatenation of previous urls with filter_params with the same
            # filter_params
            tmp_filter_params = None
            tmp_full_url = url

            if hasattr(new_records, "__iter__"):
                records.extend(new_records)
        return records

    def extract_records(
        self, url: str | None = None, report_url: str | None = None
    ) -> list[dict[str, Any]]:
        """Download records from `url` or `report_url` if present.

        Returns:
            records (List[Dict[str, Any]]): The records extracted from URL.
        """
        if self.is_report:
            return self._extract_records_from_report_url(report_url=report_url)
        full_url = urljoin(url, self.endpoint) if url else self.full_url
        return self._extract_records_from_url(url=full_url)

    def get_entities(
        self, dirty_json: dict[str, Any], url: str
    ) -> list[dict[str, Any]]:
        """Extract entities from request.json().

        Entities represent objects that store information. More info on:
        https://help.sap.com/docs/EAD_HANA/0e60f05842fd41078917822867220c78/
        0bd1db568fa546d6823d4c19a6b609ab.html

        Args:
            dirty_json (Dict[str, Any]): request.json() dict from response to API.
            url (str): The URL to fetch metadata from.

        Returns:
            entities (List[Dict[str, Any]]): list filled with entities.
        """
        metadata_url = self.create_metadata_url(url)
        column_maper_dict = self.get_property_to_sap_label_dict(metadata_url)
        entities = []
        for element in dirty_json["d"]["results"]:
            new_entity = {}
            for key, object_of_interest in element.items():
                if key not in ["__metadata", "Photo", "", "Picture"] and "{" not in str(
                    object_of_interest
                ):
                    new_key = column_maper_dict.get(key)
                    if new_key:
                        new_entity[new_key] = object_of_interest
                    else:
                        new_entity[key] = object_of_interest
            entities.append(new_entity)
        return entities

    def get_property_to_sap_label_dict(self, url: str | None = None) -> dict[str, str]:
        """Create Dict that maps Property Name to value of SAP label.

           Property: Properties define the characteristics of the data.
           SAP label: Labels are used for identification and for provision of content
           information.

        Args:
            url (str, optional): The URL to fetch metadata from.

        Returns:
            Dict[str, str]: Property Name to value of SAP label.
        """
        column_mapping = {}
        if url:
            username = self.credentials.get("username")
            password = self.credentials.get("password")
            response = requests.get(
                url,
                auth=(username, password.get_secret_value()),
                timeout=(3.05, 60 * 5),
            )
            for sentence in response.text.split("/>"):
                result = re.search(
                    r'(?<=Name=")([^"]+).+(sap:label=")([^"]+)+', sentence
                )
                if result:
                    key = result.groups(0)[0]
                    val = result.groups(0)[2]
                    column_mapping[key] = val
        return column_mapping

    def get_response(
        self,
        url: str,
        filter_params: dict[str, Any] | None = None,
        timeout: tuple = (3.05, 60 * 30),
    ) -> requests.models.Response:
        """Handle requests.

        Args:
            url (str): The url to request to.
            filter_params (Dict[str, Any], optional): Additional parameters like filter,
                used in case of normal url.
            timeout (tuple, optional): The request time-out. Default is (3.05, 60 * 30).

        Returns:
            requests.models.Response.
        """
        username = self.credentials.get("username")
        password = self.credentials.get("password")
        return handle_api_response(
            url=url,
            params=filter_params,
            auth=(username, password.get_secret_value()),
            timeout=timeout,
        )

    def to_df(
        self,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        **kwargs,
    ) -> pd.DataFrame:
        """Download a table or report into a pandas DataFrame.

        Args:
            kwargs: The parameters to pass to DataFrame constructor.

        Returns:
            df (pandas.DataFrame): DataFrame containing the records.
        """
        # Your implementation here
        if if_empty == "warn":
            self.logger.info("Warning: DataFrame is empty.")
        elif if_empty == "skip":
            self.logger.info("Skipping due to empty DataFrame.")
        elif if_empty == "fail":
            self.logger.info("Failing due to empty DataFrame.")
        else:
            msg = "Invalid value for if_empty parameter."
            raise ValueError(msg)

        url: str = kwargs.get("url", "")
        fields: list[str] = kwargs.get("fields", [])
        dtype: dict[str, Any] = kwargs.get("dtype", {})
        tests: dict[str, Any] = kwargs.get("tests", {})

        url = url or self.url
        records = self.extract_records(url=url, report_url=self.report_url)
        df = pd.DataFrame(data=records, **kwargs)

        if dtype:
            df = df.astype(dtype)

        if fields:
            return df[fields]

        if tests:
            validate(df=df, tests=tests)

        return df



from datetime import datetime, timedelta, timezone
import struct

import pytest

from viadot.sources import SQLServer


@pytest.fixture
def sql_server():
    # Initialize the SQLServer instance with the test credentials.
    return SQLServer(config_key="sql_server")


def test_handle_datetimeoffset():
    # Example test data for a datetimeoffset value.
    dto_value = struct.pack("<6hI2h", 2021, 7, 3, 14, 30, 15, 123456000, 2, 0)
    expected_datetime = datetime(
        2021, 7, 3, 14, 30, 15, 123456, tzinfo=timezone(timedelta(hours=2))
    )

    result = SQLServer._handle_datetimeoffset(dto_value)
    assert result == expected_datetime


def test_schemas(sql_server):
    schemas = sql_server.schemas
    assert "dbo" in schemas  # Assuming 'dbo' schema exists in the test database.


def test_exists(sql_server):
    sql_server_table = sql_server.tables
    sample_table_schema = sql_server_table[0].split(".")
    sample_schema = sample_table_schema[0]
    sample_table = sample_table_schema[1]
    assert sql_server.exists(table=sample_table, schema=sample_schema)



"""Exchange Rates API connector."""

from datetime import datetime
import json
from typing import Any, Literal

import pandas as pd
import requests

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.sources.base import Source
from viadot.utils import (
    add_viadot_metadata_columns,
    remove_newlines_and_tabs,
    validate,
)


Currency = Literal[
    "USD", "EUR", "GBP", "CHF", "PLN", "DKK", "COP", "CZK", "SEK", "NOK", "ISK"
]


class ExchangeRates(Source):
    URL = "https://api.apilayer.com/exchangerates_data/timeseries"

    def __init__(
        self,
        currency: Currency = "USD",
        start_date: str = datetime.today().strftime("%Y-%m-%d"),
        end_date: str = datetime.today().strftime("%Y-%m-%d"),
        symbols: list[str] | None = None,
        credentials: dict[str, Any] | None = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """Download data from https://api.apilayer.com/exchangerates_data/timeseries.

        Args:
            currency (Currency, optional): Base currency to which prices of searched
                currencies are related. Defaults to "USD".
            start_date (str, optional): Initial date for data search. Data range is
                start_date -> end_date, supported format 'yyyy-mm-dd'. Defaults to
                    datetime.today().strftime("%Y-%m-%d").
            end_date (str, optional): See above. Defaults to
                datetime.today().strftime("%Y-%m-%d").
            symbols (list, optional): List of ISO codes for which exchange rates from
                base currency will be fetched. Defaults to ["USD", "EUR", "GBP", "CHF",
                "PLN", "DKK", "COP", "CZK", "SEK", "NOK", "ISK" ].
            credentials (Dict[str, Any], optional): The credentials to use. Defaults to
                None.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials.
        """
        credentials = credentials or get_source_credentials(config_key)
        if credentials is None:
            msg = "Please specify the credentials."
            raise CredentialError(msg)
        super().__init__(*args, credentials=credentials, **kwargs)

        if not symbols:
            symbols = [
                "USD",
                "EUR",
                "GBP",
                "CHF",
                "PLN",
                "DKK",
                "COP",
                "CZK",
                "SEK",
                "NOK",
                "ISK",
            ]

        self.currency = currency
        self.start_date = start_date
        self.end_date = end_date
        self.symbols = symbols
        self._validate_symbols(self.symbols, self.currency)

    def _validate_symbols(self, symbols: list[str], currency: str):
        cur_list = [
            "USD",
            "EUR",
            "GBP",
            "CHF",
            "PLN",
            "DKK",
            "COP",
            "CZK",
            "SEK",
            "NOK",
            "ISK",
        ]

        if currency not in cur_list:
            msg = f"The specified currency does not exist or is unsupported: {currency}"
            raise ValueError(msg)

        for i in symbols:
            if i not in cur_list:
                msg = f"The specified currency list item does not exist or is not supported: {i}"
                raise ValueError(msg)

    def get_data(self) -> dict[str, Any]:
        """Download data from the API.

        Returns:
            dict[str, Any]: The data from the API.
        """
        headers = {"apikey": self.credentials["api_key"]}
        payload = {
            "start_date": self.start_date,
            "end_date": self.end_date,
            "base": self.currency,
            "symbols": ",".join(self.symbols),
        }
        response = requests.request(
            "GET", ExchangeRates.URL, headers=headers, params=payload, timeout=(3, 10)
        )

        return json.loads(response.text)

    def to_records(self) -> list[tuple]:
        """Download data and convert it to a list of records.

        Returns:
            list[tuple]: The records of the data.
        """
        data = self.get_data()
        records = []

        for j in data["rates"]:
            records.append(j)
            records.append(data["base"])

            for i in data["rates"][j]:
                records.append(data["rates"][j][i])

        return list(zip(*[iter(records)] * (2 + len(self.symbols)), strict=False))

    def get_columns(self) -> list[str]:
        """Return the columns of the data.

        Returns:
            list[str]: The columns of the data.
        """
        return ["Date", "Base", *self.symbols]

    def to_json(self) -> dict[str, Any]:
        """Download data and convert it to a JSON.

        Returns:
            dict[str, Any]: The JSON with the data.
        """
        records = self.to_records()
        columns = self.get_columns()
        records = [
            dict(zip(columns, records[i], strict=False)) for i in range(len(records))
        ]
        json = {}
        json["currencies"] = records

        return json

    @add_viadot_metadata_columns
    def to_df(
        self,
        tests: dict | None = None,
    ) -> pd.DataFrame:
        """Download data and convert it to a pandas DataFrame.

        Args:
            tests (dict | None, optional): The tests specification. Defaults to None.

        Returns:
            pd.DataFrame: The pandas DataFrame with the data.
        """
        json = self.to_json()
        df = pd.json_normalize(json["currencies"])
        df_clean = remove_newlines_and_tabs(df)

        if tests:
            validate(df=df_clean, tests=tests)

        return df_clean



from prefect import flow

from viadot.orchestration.prefect.tasks import sap_rfc_to_df


def test_sap_rfc_to_df():
    @flow
    def test_flow():
        return sap_rfc_to_df(
            config_key="sap_test",
            query="SELECT MATKL, MTART, ERSDA FROM MARA WHERE ERSDA = '20221230'",
            func="BBP_RFC_READ_TABLE",
        )

    received_df = test_flow()
    n_cols = 3

    assert not received_df.empty
    assert len(received_df.columns) == n_cols



from viadot.orchestration.prefect.flows import hubspot_to_adls


def test_hubspot_to_adls(hubspot_config_key, adls_credentials_secret):
    state = hubspot_to_adls(
        config_key=hubspot_config_key,
        endpoint="hubdb/api/v2/tables/6009756/rows/draft",
        nrows=1000000,
        adls_azure_key_vault_secret=adls_credentials_secret,
        adls_path="raw/dyvenia_sandbox/genesys/hubspot_rial.parquet",
        adls_path_overwrite=True,
    )
    assert state.is_successful()



"""Download data from Genesys Cloud and upload it to Azure Data Lake Storage."""

from typing import Any

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_adls, genesys_to_df
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="Genesys extraction to ADLS",
    description="Extract data from Genesys Cloud"
    + " and load it into Azure Data Lake Storage.",
    retries=1,
    retry_delay_seconds=60,
    log_prints=True,
)
@with_flow_timeout_param()
def genesys_to_adls(  # noqa: PLR0913
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    verbose: bool | None = None,
    endpoint: str | None = None,
    environment: str = "mypurecloud.de",
    queues_ids: list[str] | None = None,
    view_type: str | None = None,
    view_type_time_sleep: int | None = None,
    post_data_list: list[dict[str, Any]] | None = None,
    time_between_api_call: float = 0.5,
    normalization_sep: str = ".",
    drop_duplicates: bool = False,
    validate_df_dict: dict[str, Any] | None = None,
    adls_config_key: str | None = None,
    adls_azure_key_vault_secret: str | None = None,
    adls_path: str | None = None,
    adls_path_overwrite: bool = False,
) -> None:
    """Flow for downloading data from mindful to Azure Data Lake.

    Args:
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret where credentials are stored. Defaults to None.
        verbose (bool, optional): Increase the details of the logs printed on the
                screen. Defaults to False.
        endpoint (Optional[str], optional): Final end point to the API.
            Defaults to None.
        environment (str, optional): the domain that appears for Genesys Cloud
            Environment based on the location of your Genesys Cloud organization.
            Defaults to "mypurecloud.de".
        queues_ids (Optional[List[str]], optional): List of queues ids to consult the
                members. Defaults to None.
        view_type (Optional[str], optional): The type of view export job to be created.
            Defaults to None.
        view_type_time_sleep (Optional[int], optional): Waiting time to retrieve data
            from Genesys Cloud API. Defaults to None.
        post_data_list (Optional[List[Dict[str, Any]]], optional): List of string
            templates to generate json body in POST calls to the API. Defaults to None.
        time_between_api_call (int, optional): The time, in seconds, to sleep the call
            to the API. Defaults to 0.5.
        normalization_sep (str, optional): Nested records will generate names separated
            by sep. Defaults to ".".
        drop_duplicates (bool, optional): Remove duplicates from the DataFrame.
            Defaults to False.
        validate_df_dict (Optional[Dict[str, Any]], optional): A dictionary with
            optional list of tests to verify the output dataframe. Defaults to None.
        adls_config_key (Optional[str], optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        adls_azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret containing a dictionary with ACCOUNT_NAME and Service Principal
            credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
        adls_path (Optional[str], optional): Azure Data Lake destination file path (with
            file name). Defaults to None.
        adls_path_overwrite (bool, optional): Whether to overwrite the file in ADLS.
            Defaults to True.

    Examples:
        genesys_to_adls(
            config_key=config_key,
            verbose=False,
            endpoint=endpoint,
            post_data_list=data_to_post,
            adls_config_key=adls_config_key,
            adls_path=adls_path,
            adls_path_overwrite=True,
        )
    """
    data_frame = genesys_to_df(
        config_key=config_key,
        azure_key_vault_secret=azure_key_vault_secret,
        verbose=verbose,
        endpoint=endpoint,
        environment=environment,
        queues_ids=queues_ids,
        view_type=view_type,
        view_type_time_sleep=view_type_time_sleep,
        post_data_list=post_data_list,
        time_between_api_call=time_between_api_call,
        normalization_sep=normalization_sep,
        drop_duplicates=drop_duplicates,
        validate_df_dict=validate_df_dict,
    )

    df_to_adls(
        df=data_frame,
        path=adls_path,
        credentials_secret=adls_azure_key_vault_secret,
        config_key=adls_config_key,
        overwrite=adls_path_overwrite,
    )



import os

import moto
import pytest

from viadot.utils import skip_test_on_missing_extra


try:
    import boto3

    from viadot.sources import RedshiftSpectrum
except ImportError:
    skip_test_on_missing_extra("RedshiftSpectrum", extra="aws")


@pytest.fixture
def _aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"  # noqa: S105
    os.environ["AWS_SECURITY_TOKEN"] = "testing"  # noqa: S105
    os.environ["AWS_SESSION_TOKEN"] = "testing"  # noqa: S105
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"


@pytest.fixture
def _mocked_aws(_aws_credentials):
    """Mock all AWS interactions.

    Requires you to create your own boto3 clients.
    """
    with moto.mock_aws():
        yield


@pytest.fixture
def redshift_spectrum(_mocked_aws):
    conn = boto3.client("s3")
    conn.create_bucket(Bucket="test_bucket")
    spectrum = RedshiftSpectrum(config_key="redshift_dev")
    spectrum.create_schema("test_schema")

    yield spectrum

    spectrum.drop_schema("test_schema")


@pytest.mark.skip(
    reason="To be implemented: https://github.com/dyvenia/viadot/issues/978."
)
@pytest.mark.usefixtures("_mocked_aws")
def test_from_df(redshift_spectrum, TEST_DF):
    bucket = "test_bucket"
    schema = "test_schema"
    table = "test_table"
    result = redshift_spectrum.from_df(
        df=TEST_DF,
        to_path=f"s3://{bucket}/{schema}/{table}",
        schema=schema,
        table=table,
    )
    assert result is True



import pandas as pd

from viadot.orchestration.prefect.tasks import df_to_minio
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import MinIO


PATH = "data/duckdb_test.parquet"


def test_df_to_minio():
    d = {"col1": [1, 2], "col2": [3, 4]}
    df = pd.DataFrame(data=d)
    credentials = get_credentials("minio-dev")
    df_to_minio(df=df, path=PATH, credentials=credentials)
    minio = MinIO(credentials=credentials)
    file_exists = minio._check_if_file_exists(PATH)

    assert file_exists

    minio.rm(path=PATH)



"""A module for working with Azure Data Lake (gen1 and gen2)."""

from pathlib import Path
from typing import Any

import pandas as pd


try:
    from adlfs import AzureBlobFileSystem, AzureDatalakeFileSystem
except ModuleNotFoundError as e:
    msg = "Missing required modules to use AzureDataLake source."
    raise ImportError(msg) from e

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns


class AzureDataLake(Source):
    def __init__(
        self,
        path: str | None = None,
        gen: int = 2,
        credentials: dict[str, Any] | None = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """A class for pulling data from the Azure Data Lakes (gen1 and gen2).

        You can either connect to the lake in general:
        lake = AzureDataLake(); lake.exists("a/b/c.csv")

        or to a particular path:
        lake = AzureDataLake(path="a/b/c.csv"); lake.exists()`

        Args:
        credentials (Dict[str, Any], optional): A dictionary containing
            the following credentials: `account_name`, `tenant_id`,
            `client_id`, and `client_secret`.
        config_key (str, optional): The key in the viadot config holding
            relevant credentials.
        """
        credentials = credentials or get_source_credentials(config_key)
        credentials = {key.lower(): value for key, value in credentials.items()}

        required_credentials = (
            "account_name",
            "azure_tenant_id",
            "azure_client_id",
            "azure_client_secret",
        )
        required_credentials_are_provided = all(
            rc in credentials for rc in required_credentials
        )

        if credentials is None:
            msg = "Please provide the credentials."
            raise CredentialError(msg)

        if not required_credentials_are_provided:
            msg = "Please provide all required credentials."
            raise CredentialError(msg)

        super().__init__(*args, credentials=credentials, **kwargs)

        storage_account_name = self.credentials["account_name"]
        tenant_id = self.credentials["azure_tenant_id"]
        client_id = self.credentials["azure_client_id"]
        client_secret = self.credentials["azure_client_secret"]

        self.path = path
        self.gen = gen
        self.storage_options = {
            "azure_tenant_id": tenant_id,
            "azure_client_id": client_id,
            "azure_client_secret": client_secret,
        }
        if gen == 1:
            self.fs = AzureDatalakeFileSystem(
                store_name=storage_account_name,
                tenant_id=tenant_id,
                client_id=client_id,
                client_secret=client_secret,
            )
            self.base_url = f"adl://{storage_account_name}"
        elif gen == 2:  # noqa: PLR2004
            self.storage_options["account_name"] = storage_account_name
            self.fs = AzureBlobFileSystem(
                account_name=storage_account_name,
                tenant_id=tenant_id,
                client_id=client_id,
                client_secret=client_secret,
            )
            self.base_url = "az://"

    def upload(
        self,
        from_path: str,
        to_path: str | None = None,
        recursive: bool = False,
        overwrite: bool = False,
    ) -> None:
        """Upload file(s) to the lake.

        Args:
            from_path (str): Path to the local file(s) to be uploaded.
            to_path (str): Path to the destination file/folder
            recursive (bool): Set this to true if working with directories.
            overwrite (bool): Whether to overwrite the file(s) if they exist.

        Example:
        ```python
        from viadot.sources import AzureDataLake

        lake = AzureDataLake()
        lake.upload(from_path='tests/test.csv', to_path="sandbox/test.csv")
        ```
        """
        if self.gen == 1:
            msg = "Azure Data Lake Gen1 does not support simple file upload."
            raise NotImplementedError(msg)

        to_path = to_path or self.path

        self.logger.info(f"Uploading file(s) from '{from_path}' to '{to_path}'...")

        try:
            self.fs.upload(
                lpath=from_path,
                rpath=to_path,
                recursive=recursive,
                overwrite=overwrite,
            )
        except FileExistsError as e:
            if recursive:
                msg = f"At least one file in '{to_path}' already exists. Specify `overwrite=True` to overwrite."
            else:
                msg = f"The file '{to_path}' already exists. Specify `overwrite=True` to overwrite."
            raise FileExistsError(msg) from e

        self.logger.info(
            f"Successfully uploaded file(s) from '{from_path}' to '{to_path}'."
        )

    def exists(self, path: str | None = None) -> bool:
        """Check if a location exists in Azure Data Lake.

        Args:
            path (str): The path to check. Can be a file or a directory.

        Example:
        ```python
        from viadot.sources import AzureDataLake

        lake = AzureDataLake(gen=1)
        lake.exists("tests/test.csv")
        ```

        Returns:
            bool: Whether the paths exists.
        """
        path = path or self.path
        return self.fs.exists(path)

    def download(
        self,
        to_path: str,
        from_path: str | None = None,
        recursive: bool = False,
        overwrite: bool = True,
    ) -> None:
        """Download file(s) from the lake.

        Args:
            to_path (str): _description_
            from_path (str | None, optional): _description_. Defaults to None.
            recursive (bool, optional): _description_. Defaults to False.
            overwrite (bool, optional): _description_. Defaults to True.

        Raises:
            NotImplementedError: _description_
        """
        if overwrite is False:
            msg = "Currently, only the default behavior (overwrite) is available."
            raise NotImplementedError(msg)

        from_path = from_path or self.path
        self.fs.download(rpath=from_path, lpath=to_path, recursive=recursive)

    @add_viadot_metadata_columns
    def to_df(
        self,
        path: str | None = None,
        sep: str = "\t",
        quoting: int = 0,
        lineterminator: str | None = None,
        error_bad_lines: bool | None = None,
    ) -> pd.DataFrame:
        r"""Download a file from the lake and return it as a pandas DataFrame.

        Args:
            path (str, optional): _description_. Defaults to None.
            sep (str, optional): _description_. Defaults to "\t".
            quoting (int, optional): _description_. Defaults to 0.
            lineterminator (str, optional): _description_. Defaults to None.
            error_bad_lines (bool, optional): _description_. Defaults to None.

        Raises:
            ValueError: _description_

        Returns:
            _type_: _description_
        """
        if quoting is None:
            quoting = 0

        path = path or self.path
        url = self.base_url + "/" + path.strip("/")

        if url.endswith(".csv"):
            df = pd.read_csv(
                url,
                storage_options=self.storage_options,
                sep=sep,
                quoting=quoting,
                lineterminator=lineterminator,
                error_bad_lines=error_bad_lines,
            )
        elif url.endswith(".parquet"):
            df = pd.read_parquet(url, storage_options=self.storage_options)
        else:
            msg = "Only CSV and parquet formats are supported."
            raise ValueError(msg)

        return df

    def ls(self, path: str | None = None) -> list[str]:
        """Returns a list of files in a path.

        Args:
            path (str, optional): Path to a folder. Defaults to None.
        """
        path = path or self.path
        return self.fs.ls(path)

    def rm(self, path: str | None = None, recursive: bool = False) -> None:
        """Delete files in a path.

        Args:
            path (str, optional): Path to a folder. Defaults to None.
            recursive (bool, optional): Whether to delete files recursively.
                Defaults to False.
        """
        path = path or self.path
        self.fs.rm(path, recursive=recursive)

    def cp(
        self,
        from_path: str | None = None,
        to_path: str | None = None,
        recursive: bool = False,
    ) -> None:
        """Copies the contents of `from_path` to `to_path`.

        Args:
            from_path (str, optional): Path from which to copy file(s). Defaults to
                None.
            to_path (str, optional): Path where to copy file(s). Defaults to None.
            recursive (bool, optional): Whether to copy file(s) recursively. Defaults to
                False.
        """
        from_path = from_path or self.path
        self.fs.cp(from_path, to_path, recursive=recursive)

    def from_df(
        self,
        df: pd.DataFrame,
        path: str | None = None,
        sep: str = "\t",
        overwrite: bool = False,
    ) -> None:
        r"""Upload a pandas `DataFrame` to a file on Azure Data Lake.

        Args:
            df (pd.DataFrame): The pandas `DataFrame` to upload.
            path (str, optional): The destination path. Defaults to None.
            sep (str, optional): The separator to use in the `to_csv()` function.
                Defaults to "\t".
            overwrite (bool): Whether to overwrite the file if it exist.
        """
        path = path or self.path

        extension = path.split(".")[-1]
        if extension not in ("csv", "parquet"):
            if "." not in path:
                msg = "Please provide the full path to the file."
            else:
                msg = "Accepted file formats are 'csv' and 'parquet'."
            raise ValueError(msg)

        file_name = path.split("/")[-1]
        if extension == "csv":
            # Can do it simply like this if ADLS accesses are set up correctly
            # url = os.path.join(self.base_url, path)
            # df.to_csv(url, storage_options=self.storage_options)
            df.to_csv(file_name, index=False, sep=sep)
        else:
            df.to_parquet(file_name, index=False)

        self.upload(from_path=file_name, to_path=path, overwrite=overwrite)

        Path(file_name).unlink()



"""Viadot config."""

from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import Any

from yaml import safe_load


logger = logging.getLogger(__name__)

USER_HOME = Path.home()


class Config(dict):
    @classmethod
    def _get_configuration(cls, config: dict, key: str | None = None) -> dict | None:
        # Empty config.
        if not config:
            return None

        # Config key does not exist or has no values.
        if key:
            config = config.get(key)
            if not config:
                logger.warning(f"No configuration found under the '{key}' config key.")
                return None

        return cls(**config)

    @classmethod
    def from_json(cls, path: str | Path, key: str | None = None) -> Config:
        """Create a Config object from a JSON file.

        Args:
            path (str): The path to the JSON file.
            key (str | None, optional): The key inside the JSON. Defaults to None.

        Returns:
            Config: The Config object.
        """
        with Path(path).open() as f:
            config = json.load(f)
            return cls._get_configuration(config, key=key)

    @classmethod
    def from_yaml(cls, path: str | Path, key: str | None = None) -> Config:
        """Create a Config object from a YAML file.

        Args:
            path (str): The path to the YAML file.
            key (str | None, optional): The key inside the YAML. Defaults to None.

        Returns:
            Config: The Config object.
        """
        with Path(path).open() as f:
            config = safe_load(stream=f)
            return cls._get_configuration(config, key=key)


config_dir = Path(USER_HOME) / ".config" / "viadot"

try:
    CONFIG = Config.from_yaml(config_dir / "config.yaml")
except FileNotFoundError:
    try:
        CONFIG = Config.from_json(config_dir / "config.json")
    except FileNotFoundError:
        CONFIG = Config()
except ValueError:
    # Incorrect config structure.
    logger.warning(f"Could not read default viadot config file from '{config_dir}'.")
    CONFIG = Config()


def get_source_config(key: str, config: Config = CONFIG) -> dict[str, Any] | None:
    """Retrieve source configuration.

    Args:
        key (str): The key inside the config to look for.
        config (Config, optional): The config object to extract from. Defaults to
            CONFIG.

    Returns:
        dict[str, Any]: Source configuration.
    """
    source_configs = config.get("sources")
    if source_configs is not None:
        for source_config in source_configs:
            if key in source_config:
                return source_config[key]
    return None


def get_source_credentials(key: str, config: Config = CONFIG) -> dict[str, Any] | None:
    """Retrieve source credentials from the provided config.

    Args:
        key (str): The key inside the config to look for.
        config (Config, optional): The config object to extract from. Defaults to
            CONFIG.

    Returns:
        dict[str, Any]: Source credentials.
    """
    source_config = get_source_config(key, config)
    if source_config is not None:
        return source_config.get("credentials")
    return None



"""SQL Server source class."""

from datetime import datetime, timedelta, timezone
import struct

from pydantic import BaseModel, SecretStr

from viadot.config import get_source_credentials
from viadot.sources.base import SQL


class SQLServerCredentials(BaseModel):
    user: str
    password: str | SecretStr | None = None
    server: str
    driver: str = "ODBC Driver 18 for SQL Server"
    db_name: str | None = None
    trust_server_certificate: str = "yes"


class SQLServer(SQL):
    DEFAULT_SCHEMA = "dbo"

    def __init__(
        self,
        credentials: SQLServerCredentials | None = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """Connector for SQL Server.

        Args:
            credentials (SQLServerCredentials | None, optional): The credentials to use.
                Defaults to None.
            config_key (str | None, optional): The viadot config key from which to read
                the credentials. Defaults to None.
        """
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = SQLServerCredentials(**raw_creds).dict(
            by_alias=True
        )  # validate the credentials

        super().__init__(*args, credentials=validated_creds, **kwargs)
        self.server = self.credentials.get("server")
        self.username = self.credentials.get("username")
        self.password = self.credentials.get("password")
        self.driver = self.credentials.get("driver")
        self.db_name = self.credentials.get("db_name")

        self.con.add_output_converter(-155, self._handle_datetimeoffset)

    @property
    def schemas(self) -> list[str]:
        """Return a list of all schemas."""
        schemas_tuples = self.run("SELECT s.name as schema_name from sys.schemas s")
        return [schema_tuple[0] for schema_tuple in schemas_tuples]

    @property
    def tables(self) -> list[str]:
        """Return a list of all tables in the database."""
        tables_tuples = self.run(
            "SELECT schema_name(t.schema_id), t.name FROM sys.tables t"
        )
        return [".".join(row) for row in tables_tuples]

    @staticmethod
    def _handle_datetimeoffset(dto_value: str) -> datetime:
        """Adds support for SQL Server's custom `datetimeoffset` type.

        This type is not handled natively by ODBC/pyodbc.

        See: https://github.com/mkleehammer/pyodbc/issues/134#issuecomment-281739794
        """
        (
            year,
            month,
            day,
            hour,
            minute,
            second,
            nanoseconds,
            offset_hours,
            offset_minutes,
        ) = struct.unpack("<6hI2h", dto_value)
        return datetime(
            year,
            month,
            day,
            hour,
            minute,
            second,
            nanoseconds // 1000,
            tzinfo=timezone(timedelta(hours=offset_hours, minutes=offset_minutes)),
        )

    def exists(self, table: str, schema: str | None = None) -> bool:
        """Check whether a table exists.

        Args:
            table (str): The table to be checked.
            schema (str, optional): The schema where the table is located.
                Defaults to 'dbo'.

        Returns:
            bool: Whether the table exists.
        """
        if not schema:
            schema = self.DEFAULT_SCHEMA

        list_table_info_query = f"""
            SELECT *
            FROM sys.tables t
            JOIN sys.schemas s
                ON t.schema_id = s.schema_id
            WHERE s.name = '{schema}' AND t.name = '{table}'
        """  # noqa: S608
        return bool(self.run(list_table_info_query))



"""Viadot exceptions."""


class ValidationError(Exception):
    pass


class APIError(Exception):
    pass


class CredentialError(Exception):
    pass


class DBDataAccessError(Exception):
    pass


class DataRangeError(Exception):
    pass


class TableDoesNotExistError(Exception):
    def __init__(
        self,
        message: str | None = None,
        table: str | None = None,
        schema: str | None = None,
        fqn: str | None = None,
    ):
        """Raise when a database table does not exist.

        Args:
            table (str, optional): The name of the table. Defaults to None.
            schema (str, optional): The schema where the table is located. Defaults to
                None.
            fqn (str, optional): The fully-qualified name of the table. Defaults to
                None.
        """
        fqn = f"{schema}.{table}" if table and schema else fqn or table

        message = message or f"Table {fqn} does not exist."

        super().__init__(message)


class TableAlreadyExistsError(Exception):
    def __init__(
        self,
        message: str | None = None,
        table: str | None = None,
        schema: str | None = None,
        fqn: str | None = None,
    ):
        """Raise when a database table already exists.

        Args:
            table (str, optional): The name of the table. Defaults to None.
            schema (str, optional): The schema where the table is located. Defaults to
                None.
            fqn (str, optional): The fully-qualified name of the table. Defaults to
                None.
        """
        fqn = f"{schema}.{table}" if table and schema else fqn or table

        message = message or f"Table {fqn} already exists."

        super().__init__(message)


class DataBufferExceededError(Exception):
    pass


class NoFilesToProcessError(Exception):
    pass


class SMBError(Exception):
    pass


class SMBConnectionError(SMBError):
    pass


class SMBFileOperationError(SMBError):
    pass


class SMBInvalidFilenameError(SMBError):
    pass



"""OneStream API connector for data extraction and management.

This module provides the OneStream class for interacting with OneStream API endpoints,
including Data Adapter queries, SQL queries, and Data Management sequence execution.
"""

from itertools import product
import json
from typing import Any, Literal

import pandas as pd
from pydantic import BaseModel
import requests

from viadot.config import get_source_credentials
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns, handle_api_request


class OneStreamCredentials(BaseModel):
    """Check OneStream credentials dict.

    OneStream API Key token is required:
        - api_token: The unique authentication api token for the OneStream API.

    Args:
         BaseModel (pydantic.BaseModel): Base class for data validation.
    """

    api_token: str


class OneStream(Source):
    def __init__(
        self,
        base_url: str,
        application: str,
        config_key: str | None = None,
        credentials: OneStreamCredentials | None = None,
        params: dict[str, str] | None = None,
        *args,
        **kwargs,
    ):
        """Connector class to ingest data from the OneStream API endpoint.

        API Documentation:
        https://documentation.onestream.com/1388457/Content/REST%20API/OneStream%20WebAPI%20Endpoints.html

        Provides access to various OneStream API endpoints for data extraction and
        management. Each endpoint requires a specific function designed to handle its
        unique requirements.

        Args:
            base_url (str): OneStream base server URL.
            application (str): Name of the OneStream application to connect to.
            config_key (str, optional): Key in viadot config to fetch credentials.
                Defaults to None.
            credentials (OneStreamCredentials, optional): OneStream API credentials.
                Must contain 'api_token'. If not provided, will attempt to fetch from
                viadot config using config_key. Defaults to None.
            params (dict[str, str], optional): Additional API parameters to include
                in requests. Defaults to None, which sets {"api-version": "5.2.0"}.
            *args: Additional positional arguments passed to parent class.
            **kwargs: Additional keyword arguments passed to parent class.

        Note:
            The connector supports three main operations:
                1. Data Adapter queries: Fetch data using OneStream Data Adapters.
                2. SQL queries: Execute SQL queries against OneStream databases.
                3. Data Management: Run Data Management sequences.
        """
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = dict(OneStreamCredentials(**raw_creds))
        super().__init__(*args, credentials=validated_creds, **kwargs)
        self.base_url = base_url
        self.application = application
        self.api_token = self.credentials.get("api_token")
        params_with_default_api_version = {"api-version": "5.2.0"}
        params_with_default_api_version.update(params or {})
        self.params = params_with_default_api_version

    def _execute_api_method(
        self,
        api: Literal["data_adapter", "sql_query", "run_data_management_seq"],
        **kwargs,
    ) -> list[dict[str, Any]] | requests.Response:
        """Executes function for selected api endpoint.

        Args:
            api: The API endpoint type to execute.
            **kwargs: Additional arguments passed to the specific API method.

        Returns:
            list[dict[str, Any]] | requests.Response: The result of the API call.
        """
        match api:
            case "data_adapter":
                return self._get_agg_data_adapter_endpoint_data(**kwargs)
            case "sql_query":
                return self._get_agg_sql_query_endpoint_data(**kwargs)
            case "run_data_management_seq":
                return self._run_data_management_seq(**kwargs)

    def _extract_data_from_response(
        self, response: requests.Response, adapter_response_key: str
    ) -> dict[str, Any]:
        """Extract records from the API response.

        Filters API response based on the adapter response key.

        Args:
            response (requests.Response): The HTTP response object from the API call.
            adapter_response_key (str): The key in the JSON response that contains
                the adapter's returned records.

        Returns:
            dict[str, Any]: The extracted records from the JSON response under
                the specified key.

        Raises:
            ValueError: If the API response is null or the key is not found.
        """
        json_response = response.json()

        if json_response is None:
            self.logger.error(
                "API call returned null response. Consider adding additional substition custom variables."
            )
            msg = "API response is null."
            raise ValueError(msg)
        return json_response[adapter_response_key]

    def _unpack_custom_subst_vars_to_string(
        self, custom_subst_vars: dict[str, Any]
    ) -> str:
        """Converts a dictionary into a comma-separated string of key=value pairs.

        Args:
            custom_subst_vars (dict[str, Any]): A dictionary mapping substitution
                variable names to lists of possible values.

        Returns:
            str: A key-value pair string formatted as 'key=value', separated by commas.
        """
        custom_subst_vars_list = [
            f"{key}={element}" for key, element in custom_subst_vars.items()
        ]
        return ",".join(custom_subst_vars_list)

    def _get_all_custom_subst_vars_combinations(
        self, custom_subst_vars: dict[str, list[Any]]
    ) -> list[dict[str, Any]]:
        """Generates a list of dictionaries of all combinations of custom variables.

        Args:
            custom_subst_vars (dict[str, list[Any]]): A dictionary where each key
                maps to a list of possible values for that variable. The
                cartesian product of these lists will be computed.

        Returns:
            list[dict[str, Any]]: A list of dictionaries, each representing
                a unique combination of custom variable values, where each
                dictionary has the same keys as the input but with single
                values selected from the corresponding lists.
        """
        return [
            dict(zip(custom_subst_vars.keys(), combination, strict=False))
            for combination in product(*custom_subst_vars.values())
        ]

    def _get_agg_data_adapter_endpoint_data(
        self,
        adapter_name: str,
        workspace_name: str = "MainWorkspace",
        adapter_response_key: str = "Results",
        custom_subst_vars: dict[str, list[Any]] | None = None,
    ) -> list[dict[str, Any]]:
        """Fetch and aggregate data from a OneStream Data Adapter (DA).

        This function constructs API requests (one per substitution-variable
        combination, if provided) and aggregates the results into a list of
        datasets.

        Args:
            adapter_name (str): The name of the Data Adapter to query.
            workspace_name (str, optional): The workspace name where the DA
                is located. Defaults to "MainWorkspace".
            adapter_response_key (str, optional): The key in the JSON
                response that contains the adapter's returned data.
                Defaults to "Results".
            custom_subst_vars (dict[str, list[Any]], optional): A dictionary
                mapping substitution variable names to lists of possible values.
                Defaults to None.

        Returns:
            list[dict[str, Any]]: A flat list of all record dictionaries
                aggregated from all substitution variable combinations.
        """
        custom_subst_vars = custom_subst_vars or {}
        custom_subst_vars_list = self._get_all_custom_subst_vars_combinations(
            custom_subst_vars
        )

        agg_records = []
        for custom_subst_vars in custom_subst_vars_list:
            agg_records.append(
                self._fetch_adapter_results_data(
                    workspace_name=workspace_name,
                    adapter_response_key=adapter_response_key,
                    custom_subst_vars=custom_subst_vars,
                    adapter_name=adapter_name,
                )
            )

        return agg_records

    def _fetch_adapter_results_data(
        self,
        adapter_name: str,
        workspace_name: str,
        adapter_response_key: str,
        custom_subst_vars: dict[str, Any],
    ) -> list[dict[str, Any]]:
        """Fetch data from a specified Data Adapter (DA) in OneStream.

        Makes a direct API call to fetch data from a Data Adapter in the
        specified workspace. The method constructs a POST request with the
        adapter details and custom substitution variables, then extracts the
        relevant data from the response.

        Args:
            adapter_name (str): The name of the Data Adapter to query.
            workspace_name (str): The name of the workspace where the DA
                is located.
            adapter_response_key (str): The key in the JSON response that
                contains the adapter's returned data.
            custom_subst_vars (dict[str, Any]):A dictionary mapping substitution
                variable names to lists of possible values.

        Returns:
            list[dict[str, Any]]: A list of record dictionaries from the
                adapter response.

        Raises:
            ValueError: If the API response is null or missing the
                specified response key.
            requests.exceptions.RequestException: If the API request fails
                or times out.
        """
        custom_subst_vars_str = self._unpack_custom_subst_vars_to_string(
            custom_subst_vars
        )
        endpoint = (
            self.base_url.rstrip("/") + "/api/DataProvider/GetAdoDataSetForAdapter"
        )

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_token}",
            "Accept-Encoding": "gzip, deflate, br",
        }
        payload = json.dumps(
            {
                "BaseWebServerUrl": self.base_url.rstrip("/") + "/OneStreamWeb",
                "ApplicationName": self.application,
                "WorkspaceName": workspace_name,
                "AdapterName": adapter_name,
                "ResultDataTableName": adapter_response_key,
                "CustomSubstVarsAsCommaSeparatedPairs": custom_subst_vars_str,
            }
        )

        response = handle_api_request(
            method="POST",
            url=endpoint,
            params=self.params,
            headers=headers,
            data=payload,
            timeout=(60, 3600),
        )

        return self._extract_data_from_response(response, adapter_response_key)

    def _get_agg_sql_query_endpoint_data(
        self,
        custom_subst_vars: dict | None = None,
        sql_query: str = "",
        db_location: str = "Application",
        results_table_name: str = "Results",
        external_db: str = "",
    ) -> list[dict[str, Any]]:
        """Fetch and aggregate data from a SQL query endpoint in OneStream.

        This function constructs and executes SQL queries using the provided
        parameters and custom variable combinations. It is typically used to
        retrieve configuration-level data such as metadata, security groups,
        or settings.

        Args:
            sql_query (str): The SQL query to execute.
            db_location (str, optional): The database context in which to run
                the query. Try "Framework" for system level config data.
                Defaults to "Application".
            results_table_name (str, optional): The key in the JSON response
                containing the desired data. Defaults to "Results".
            external_db (str, optional): The name of an external database.
                Defaults to an empty string.
            custom_subst_vars (dict[str, list[Any]], optional): A dictionary
                mapping substitution variable names to lists of possible values.
                Defaults to None.

        Returns:
            list[dict[str, Any]]: A flat list of all record dictionaries
                aggregated from all substitution variable combinations.
        """
        custom_subst_vars = custom_subst_vars or {}

        custom_subst_vars_list = self._get_all_custom_subst_vars_combinations(
            custom_subst_vars
        )

        agg_records = []

        for custom_subst_var in custom_subst_vars_list:
            agg_records.append(
                self._run_sql(
                    sql_query=sql_query,
                    db_location=db_location,
                    results_table_name=results_table_name,
                    external_db=external_db,
                    custom_subst_vars=custom_subst_var,
                )
            )

        return agg_records

    def _run_sql(
        self,
        sql_query: str,
        db_location: str,
        results_table_name: str,
        external_db: str,
        custom_subst_vars: dict[str, list[Any]] | None = None,
    ) -> dict[str, Any]:
        """Execute an SQL query against the OneStream database using the API.

        Constructs and sends a POST request to the OneStream API endpoint
        with the provided SQL query parameters. The response is parsed to
        extract the results.

        Args:
            sql_query (str): SQL query to execute.
            db_location (str): Target database location within OneStream.
            results_table_name (str): Name of the results table.
            external_db (str): Name of external database connection.
            custom_subst_vars (dict[str, list[Any]], optional): A dictionary mapping
                substitution variable names to lists of possible values.
                Defaults to None.

        Returns:
            dict[str, Any]: The extracted records from the specified
                results table.

        Raises:
            ValueError: If the API response is null or the results table
                is not found.
        """
        custom_subst_vars = custom_subst_vars or {}
        custom_subst_vars_str = self._unpack_custom_subst_vars_to_string(
            custom_subst_vars
        )

        endpoint = (
            self.base_url.rstrip("/") + "/api/DataProvider/GetAdoDataSetForSqlCommand"
        )

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_token}",
        }
        payload = json.dumps(
            {
                "BaseWebServerUrl": self.base_url.rstrip("/") + "/OneStreamWeb",
                "ApplicationName": self.application,
                "SqlQuery": sql_query,
                "ResultDataTableName": results_table_name,
                "DbLocation": db_location,
                "XFExternalDBConnectionName": external_db,
                "CustomSubstVarsAsCommaSeparatedPairs": custom_subst_vars_str,
            }
        )

        response = handle_api_request(
            method="POST",
            url=endpoint,
            params=self.params,
            headers=headers,
            data=payload,
            timeout=(60, 3600),
        )
        return self._extract_data_from_response(response, results_table_name)

    def _run_data_management_seq(
        self,
        dm_seq_name: str,
        custom_subst_vars: dict | None = None,
    ) -> requests.Response:
        """Executes a Data Management (DM) sequence in OneStream using the API.

        Constructs and sends a POST request to the OneStream DataManagement API to run
        the specified sequence, optionally including custom substitution variables.

        Args:
            dm_seq_name (str): The name of the Data Management sequence to execute.
            custom_subst_vars (dict[str, Any], optional): A dictionary mapping
                substitution variable names to lists of possible values
                to pass to the sequence. Defaults to None.

        Returns:
            requests.Response: The HTTP response object returned by the API after
                executing the sequence.
        """
        custom_subst_vars = custom_subst_vars or {}
        custom_subst_vars_str = self._unpack_custom_subst_vars_to_string(
            custom_subst_vars
        )

        endpoint = self.base_url.rstrip("/") + "/api/DataManagement/ExecuteSequence"

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_token}",
        }

        payload = json.dumps(
            {
                "BaseWebServerUrl": self.base_url.rstrip("/") + "/OneStreamWeb",
                "ApplicationName": self.application,
                "SequenceName": dm_seq_name,
                "CustomSubstVarsAsCommaSeparatedPairs": custom_subst_vars_str,
            }
        )

        return handle_api_request(
            method="POST",
            url=endpoint,
            params=self.params,
            headers=headers,
            data=payload,
            timeout=(60, 3600),
        )

    @add_viadot_metadata_columns
    def to_df(
        self,
        api: Literal["data_adapter", "sql_query"],
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        **kwargs,
    ) -> pd.DataFrame:
        """Get data from the API endpoint and convert to a Pandas DataFrame.

        This method fetches data from the specified API endpoint and converts
        it to a DataFrame. All endpoint-specific parameters should be passed
        as keyword arguments.

        Args:
            api (Literal["data_adapter", "sql_query"]): The API endpoint type
                to query. Only endpoints that return tabular data are supported.
            if_empty (Literal["warn", "skip", "fail"], optional): Action to
                take if the DataFrame is empty.
                - "warn": Logs a warning.
                - "skip": Skips the operation.
                - "fail": Raises an error.
                Defaults to "warn".
            **kwargs: Endpoint-specific parameters passed to the underlying
                API method.
                For "data_adapter":
                    - adapter_name (str): Name of the Data Adapter to query.
                    - workspace_name (str, optional): Workspace name.
                      Defaults to "MainWorkspace".
                    - adapter_response_key (str, optional): Response key.
                      Defaults to "Results".
                    - custom_subst_vars (dict[str, list[Any]], optional):
                      Custom substitution variables.
                For "sql_query":
                    - sql_query (str): The SQL query to execute.
                    - custom_subst_vars (dict, optional): Custom
                      substitution variables.
                    - db_location (str, optional): Database location.
                      Defaults to "Application".
                    - results_table_name (str, optional): Results table name.
                      Defaults to "Results".
                    - external_db (str, optional): External database name.
                      Defaults to "".

        Returns:
            pd.DataFrame: A DataFrame containing all records from the fetched
                data.
        """
        data = self._execute_api_method(api=api, **kwargs)
        unpacked_data = [part for parts in data for part in parts]
        df = pd.DataFrame(unpacked_data)
        if df.empty:
            msg = "The response data is empty."
            self._handle_if_empty(if_empty, message=msg)
        return df

    def execute(
        self,
        api: Literal["run_data_management_seq"],
        **kwargs,
    ) -> requests.Response:
        """Execute an API endpoint operation that returns a Response object.

        This method is for API endpoints that don't return tabular data
        (e.g., "run_data_management_seq"). For data extraction endpoints,
        use to_df() instead. All endpoint-specific parameters should be
        passed as keyword arguments.

        Args:
            api (Literal["run_data_management_seq"]): The API endpoint type
                to execute. Only endpoints that don't return tabular data
                are supported.
            **kwargs: Endpoint-specific parameters.
                For "run_data_management_seq":
                    - dm_seq_name (str): The name of the Data Management
                      sequence to execute.
                    - custom_subst_vars (dict, optional): Custom substitution
                      variables.

        Returns:
            requests.Response: The HTTP response object from the API.
        """
        return self._execute_api_method(api=api, **kwargs)



"""Extract data from OneStream Data Adapters and load it to Redshift Spectrum."""

from typing import Any, Literal

from prefect import flow
from prefect.logging import get_run_logger

from viadot.orchestration.prefect.tasks import (
    create_batch_list_of_custom_subst_vars,
    df_to_redshift_spectrum,
    onestream_to_df,
)
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="extract--onestream_data_adapters--redshift_spectrum",
    description="Extract data from OneStream Data Adapters and load it to Redshift Spectrum.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def onestream_data_adapters_to_redshift_spectrum(  # noqa: PLR0913
    base_url: str,
    application: str,
    adapter_name: str,
    to_path: str,
    schema_name: str,
    table: str,
    workspace_name: str = "MainWorkspace",
    adapter_response_key: str = "Results",
    custom_subst_vars: dict[str, list[Any]] | None = None,
    batch_by_subst_vars: bool = False,
    params: dict[str, str] | None = None,
    extension: str = ".parquet",
    if_exists: Literal["overwrite", "append"] = "overwrite",
    if_empty: Literal["warn", "skip", "fail"] = "fail",
    partition_cols: list[str] | None = None,
    index: bool = False,
    compression: Literal["snappy", "gzip", "zstd"] | None = None,
    sep: str = ",",
    aws_config_key: str | None = None,
    credentials_secret: str | None = None,
    onestream_credentials_secret: str | None = None,
    onestream_config_key: str | None = None,
) -> None:
    """Extract data from OneStream Data Adapter and load it into AWS Redshift Spectrum.

    API documentation:
    https://documentation.onestream.com/1384528/Content/REST%20API/OneStream%20WebAPI%20Endpoints.html#:~:text=Data%20Provider%20GetAdoDataSetForAdapter%20endpoint

    This flow retrieves data from a OneStream Data Adapter using provided parameters
    and uploads it to AWS Redshift Spectrum.

    Flow can be executed with the custom substitution variables mapped to its values.
    These variable-value combination is used to filter data inside the data
    adapters table allowing to return only the datasets where rows have
    the value for a given column values set.The column - variable mapping is defined in
    the OneStream Data Adapter configuration.The filtering is done for each
    variable-values combination, hence the flow might generate a couple of API calls
    to get data for all combinations.

    When custom_subst_vars are provided and batch_by_subst_vars is True, the ingestion
    process will be split into batches. Each batch represents one combination of the
    custom_subst_vars - variable:[value]. When batch_by_subst_vars is False, all
    substitution variable combinations are processed together into a single data frame.
    Warning! Processing custom substitution vars without batching might lead to
    out-of-memory errors when each variable-value combination data set size is too big.

    Args:
        base_url (str): OneStream base server URL.
        application (str): OneStream application name.
        adapter_name (str): Data Adapter name to query.
        to_path (str): Path to an S3 folder where the table will be located.
        schema_name (str): AWS Glue catalog database name.
        table (str): AWS Glue catalog table name.
        workspace_name (str): OneStream workspace name. Defaults to "MainWorkspace".
        adapter_response_key (str): Key in the JSON response that contains
            the adapter's returned data. Defaults to "Results".
        custom_subst_vars (dict[str, list[Any]], optional): A dictionary mapping
            substitution variable names to lists of possible values.Substition variables
            used as substitution variables that are mapped to the column names in the
            Data Adapter configuration.They provides a kind of SQL WHERE clause for rows
            filtering where variable refers to a name of a set containings a list
            of possible values.
            For example: "prm_activity_region":["MK", "LG"]
                        refers to column UD3 that has custom variable mapping of
                        prm_activity_region.Data will be extracted only for rows where
                        the values (MK,LG) are present.
            Values can be of any type that can be converted to strings.Defaults to None.
        batch_by_subst_vars (bool): Whether to process data in batches based on
            substitution variable combinations. When True and custom_subst_vars is
            provided, each combination of substitution variables and its values will be
            processed as a separate batch, uploading individual parquet files in S3.
            When False,all substitution variable combinations are collected in a loop
            and are uploaded together in a single operation. Defaults to False.
        params (dict[str, str], optional): Additional API parameters. Defaults to None.
        extension (str): Required file type. Accepted formats: 'csv', 'parquet'.
            Defaults to ".parquet".
        if_exists (Literal["overwrite", "append"], optional): Whether to 'overwrite'
            or 'append' to existing table. Defaults to "overwrite".
        if_empty (Literal["warn", "skip", "fail"], optional): What to do if the
            API returns no data. Defaults to "fail".
        partition_cols (list[str], optional): Columns used to create partitions.
            Only applies when dataset=True. Defaults to None.
        index (bool): Write row names (index). Defaults to False.
        compression (Literal["snappy", "gzip", "zstd"], optional): Compression style
            (None, snappy, gzip, zstd). Defaults to None.
        sep (str): Field delimiter for the output file. Defaults to ','.
        aws_config_key (str, optional): Key in viadot config for AWS credentials.
            Defaults to None.
        credentials_secret (str, optional): Name of Prefect secret block with AWS
            credentials. Defaults to None.
        onestream_credentials_secret (str, optional): Name of secret storing OneStream
            credentials. Defaults to None.
        onestream_config_key (str, optional)): Key in viadot config for OneStream
            credentials.Defaults to None.
    """
    logger = get_run_logger()

    logger.info(f"Starting OneStream Data Adapter extraction: {adapter_name}")

    if batch_by_subst_vars and not custom_subst_vars:
        msg = (
            "Invalid parameter combination: batch_by_subst_vars=True requires "
            "custom_subst_vars to be provided. Either set batch_by_subst_vars=False "
            "or provide a custom_subst_vars dictionary."
        )
        logger.error(msg)
        raise ValueError(msg)
    if custom_subst_vars and batch_by_subst_vars:
        # Process data in batches - each substitution variable combination separately
        custom_subst_vars_batch_list = create_batch_list_of_custom_subst_vars(
            custom_subst_vars
        )
        logger.info(
            f"Processing {len(custom_subst_vars_batch_list)} batches based on substitution variable combinations."
        )

        for i, custom_subst_var in enumerate(custom_subst_vars_batch_list, 1):
            logger.info(
                f"Processing batch {i}/{len(custom_subst_vars_batch_list)}: {custom_subst_var}."
            )
            df = onestream_to_df(
                base_url=base_url,
                application=application,
                api="data_adapter",
                adapter_name=adapter_name,
                workspace_name=workspace_name,
                adapter_response_key=adapter_response_key,
                custom_subst_vars=custom_subst_var,
                if_empty=if_empty,
                params=params,
                credentials_secret=onestream_credentials_secret,
                config_key=onestream_config_key,
            )
            df_to_redshift_spectrum(
                df=df,
                to_path=to_path,
                schema_name=schema_name,
                table=table,
                extension=extension,
                if_exists=if_exists,
                partition_cols=partition_cols,
                index=index,
                compression=compression,
                sep=sep,
                config_key=aws_config_key,
                credentials_secret=credentials_secret,
            )
            if_exists = "append"  # Changed to "append" to add batches to the table.
            logger.info(
                f"Batch {i}/{len(custom_subst_vars_batch_list)} completed successfully."
            )

        logger.info("All batches processed successfully.")
    else:
        # Process all data together - either no custom_subst_vars or batching disabled
        if custom_subst_vars:
            logger.info(
                "Processing all substitution variable combinations as one data frame."
            )
        else:
            logger.info("Processing data without substitution variables.")
        df = onestream_to_df(
            base_url=base_url,
            application=application,
            api="data_adapter",
            adapter_name=adapter_name,
            workspace_name=workspace_name,
            adapter_response_key=adapter_response_key,
            custom_subst_vars=custom_subst_vars,
            params=params,
            credentials_secret=onestream_credentials_secret,
            config_key=onestream_config_key,
        )
        df_to_redshift_spectrum(
            df=df,
            to_path=to_path,
            schema_name=schema_name,
            table=table,
            extension=extension,
            if_exists=if_exists,
            partition_cols=partition_cols,
            index=index,
            compression=compression,
            sep=sep,
            config_key=aws_config_key,
            credentials_secret=credentials_secret,
        )
        logger.info("Data processing completed successfully.")



"""Task for uploading pandas DataFrame to AWS Redshift Spectrum."""

import contextlib
from typing import Any, Literal

import pandas as pd
from prefect import task
from prefect.logging import get_run_logger

from viadot.config import get_source_credentials
from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials


with contextlib.suppress(ImportError):
    from viadot.sources import RedshiftSpectrum


@task(
    retries=3,
    retry_delay_seconds=10,
    timeout_seconds=60 * 60,
    cache_result_in_memory=False,
)
def df_to_redshift_spectrum(  # noqa: PLR0913
    df: pd.DataFrame,
    to_path: str,
    schema_name: str,
    table: str,
    extension: str = ".parquet",
    if_exists: Literal["overwrite", "append"] = "overwrite",
    partition_cols: list[str] | None = None,
    index: bool = False,
    compression: str | None = None,
    sep: str = ",",
    config_key: str | None = None,
    credentials_secret: str | None = None,
    **kwargs: dict[str, Any] | None,
) -> None:
    """Task to upload a pandas `DataFrame` to a csv or parquet file.

    Args:
        df (pd.DataFrame): The Pandas DataFrame to ingest into Redshift Spectrum.
        to_path (str): Path to a S3 folder where the table will be located. If needed,
            a bottom-level directory named f"{table}" is automatically created, so
            that files are always located in a folder named the same as the table.
        schema_name (str): AWS Glue catalog database name.
        table (str): AWS Glue catalog table name.
        partition_cols (list[str]): List of column names that will be used to create
            partitions. Only takes effect if dataset=True.
        extension (str): Required file type. Accepted file formats are 'csv' and
            'parquet'.
        if_exists (str, optional): 'overwrite' to recreate any possible existing table
            or 'append' to keep any possible existing table. Defaults to overwrite.
        partition_cols (list[str], optional): List of column names that will be used to
            create partitions. Only takes effect if dataset=True. Defaults to None.
        index (bool, optional): Write row names (index). Defaults to False.
        compression (str, optional): Compression style (None, snappy, gzip, zstd).
        sep (str, optional): Field delimiter for the output file. Applies only to '.csv'
            extension. Defaults to ','.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        credentials_secret (str, optional): The name of a secret block in Prefect
            that stores AWS credentials. Defaults to None.
        kwargs: The parameters to pass in awswrangler to_parquet/to_csv function.
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    credentials = get_source_credentials(config_key) or get_credentials(
        credentials_secret
    )

    # Convert columns containing only null values to string to avoid errors
    # during new table creation
    null_columns = df.isna().all()
    null_cols_list = null_columns[null_columns].index.tolist()
    df[null_cols_list] = df[null_cols_list].astype("string")

    rs = RedshiftSpectrum(credentials=credentials, config_key=config_key)

    rs.from_df(
        df=df,
        to_path=to_path,
        schema=schema_name,
        table=table,
        extension=extension,
        if_exists=if_exists,
        partition_cols=partition_cols,
        index=index,
        compression=compression,
        sep=sep,
        **kwargs,
    )

    logger = get_run_logger()
    logger.info("Data has been uploaded successfully.")



"""The viadot library."""

import logging


# Remove trash Azure INFO logs which contain low-level debugging information
# but keep WARNING and higher ones in case something actually important happens.

azure_clutter_logger_1 = logging.getLogger(
    "azure.core.pipeline.policies.http_logging_policy",
)
azure_clutter_logger_1.setLevel(logging.WARNING)


azure_clutter_logger_2 = logging.getLogger(
    "azure.identity.aio._internal.get_token_mixin",
)
azure_clutter_logger_2.setLevel(logging.WARNING)



import os
from pathlib import Path

import pytest

from viadot.exceptions import CredentialError
from viadot.sources import Sharepoint


def test_is_configured():
    s = Sharepoint(
        credentials={
            "site": "test_site",
            "username": "test_user",
            "password": "test_password",
        },
    )
    assert s


def test_is_configured_throws_credential_error():
    with pytest.raises(CredentialError):
        _ = Sharepoint(
            credentials={
                "site": None,
                "username": "test_user",
                "password": "test_password",
            },
        )
    with pytest.raises(CredentialError):
        _ = Sharepoint(
            credentials={
                "site": "test_site",
                "username": None,
                "password": "test_password",
            },
        )
    with pytest.raises(CredentialError):
        _ = Sharepoint(
            credentials={
                "site": "test_site",
                "username": "test_user",
                "password": None,
            },
        )


def test_to_df(sharepoint_url, sharepoint_config_key):
    s = Sharepoint(config_key=sharepoint_config_key)
    df = s.to_df(sharepoint_url, usecols="A")
    assert not df.empty


def test_download_file(sharepoint_url, sharepoint_config_key):
    s = Sharepoint(config_key=sharepoint_config_key)

    file = "sharepoint_test" + "." + sharepoint_url.split(".")[-1]
    s.download_file(url=sharepoint_url, to_path=file)

    assert file in os.listdir()

    Path(file).unlink()



import os

from viadot.orchestration.prefect.flows import sharepoint_to_redshift_spectrum
from viadot.sources import RedshiftSpectrum


def test_sharepoint_to_redshift_spectrum(sharepoint_url, sharepoint_credentials_secret):
    s3_bucket = os.environ.get("S3_BUCKET")
    test_schema = "raw_test"
    test_table = "test_sharepoint_to_redshift_spectrum"

    sharepoint_to_redshift_spectrum(
        sharepoint_url=sharepoint_url,
        to_path=f"s3://{s3_bucket}/nesso/{test_schema}/{test_table}",
        schema_name=test_schema,
        table=test_schema,
        sharepoint_credentials_secret=sharepoint_credentials_secret,
        download_all_files=False,
        return_as_one_table=False,
        file_sheet_mapping={},
    )

    rs = RedshiftSpectrum()

    received_df = rs.to_df(database=test_schema, table=test_table)

    assert not received_df.empty

    rs.rm(database=test_schema, table=test_table)



"""A connector for Azure SQL Database."""

import logging
from typing import Literal

import pandas as pd

from viadot.utils import df_clean_column, df_converts_bytes_to_int

from .sql_server import SQLServer


logger = logging.getLogger(__name__)


class AzureSQL(SQLServer):
    """Azure SQL connector class."""

    def __init__(self, *args, config_key: str = "AZURE_SQL", **kwargs):
        """Initialize the AzureSQL connector.

        This constructor sets up the Azure SQL connector with the specified
        configuration key. It allows for additional positional and keyword arguments
        to be passed to the parent SQLServer class.

        Args:
            *args: Variable length argument list passed to the parent class.
            config_key (str, optional): The configuration key used to retrieve
                connection settings. Defaults to "AZURE_SQL".
            **kwargs: Additional keyword arguments passed to the parent class.
        """
        super().__init__(*args, config_key=config_key, **kwargs)

    def bulk_insert(
        self,
        table: str,
        schema: str | None = None,
        source_path: str | None = None,
        sep: str | None = "\t",
        if_exists: Literal["append", "replace"] = "append",
    ) -> bool:
        r"""Function to bulk insert.

        Args:
            table (str): Table name.
            schema (str, optional): Schema name. Defaults to None.
            source_path (str, optional): Full path to a data file. Defaults to one.
            sep (str, optional):  field terminator to be used for char and
                widechar data files. Defaults to "\t".
            if_exists (Literal["append", "replace"] , optional): What to do if the table
                already exists. Defaults to "append".
        """
        if schema is None:
            schema = self.DEFAULT_SCHEMA
        fqn = f"{schema}.{table}"
        insert_sql = f"""
            BULK INSERT {fqn} FROM '{source_path}'
            WITH (
                CHECK_CONSTRAINTS,
                DATA_SOURCE='{self.credentials['data_source']}',
                DATAFILETYPE='char',
                FIELDTERMINATOR='{sep}',
                ROWTERMINATOR='0x0a',
                FIRSTROW=2,
                KEEPIDENTITY,
                TABLOCK,
                CODEPAGE='65001'
            );
        """
        if if_exists == "replace":
            self.run(f"DELETE FROM {schema}.{table}")  # noqa: S608
        self.run(insert_sql)
        return True

    def create_external_database(
        self,
        external_database_name: str,
        storage_account_name: str,
        container_name: str,
        sas_token: str,
        master_key_password: str,
        credential_name: str | None = None,
    ) -> None:
        """Create an external database.

        Used to eg. execute BULK INSERT or OPENROWSET queries.

        Args:
            external_database_name (str): The name of the extrnal source (db)
                to be created.
            storage_account_name (str): The name of the Azure storage account.
            container_name (str): The name of the container which should
                become the "database".
            sas_token (str): The SAS token to be used as the credential.
                Note that the auth system in Azure is pretty broken and you might need
                to paste here your storage account's account key instead.
            master_key_password (str): The password for the database master key of your
                Azure SQL Database.
            credential_name (str): How to name the SAS credential. This is really
                an Azure internal thing and can be anything.
                By default '{external_database_name}_credential`.
        """
        # stupid Microsoft thing
        if sas_token.startswith("?"):
            sas_token = sas_token[1:]

        if credential_name is None:
            credential_name = f"{external_database_name}_credential"

        create_master_key_sql = (
            f"CREATE MASTER KEY ENCRYPTION BY PASSWORD = {master_key_password}"
        )

        create_external_db_credential_sql = f"""
        CREATE DATABASE SCOPED CREDENTIAL {credential_name}
        WITH IDENTITY = 'SHARED ACCESS SIGNATURE'
        SECRET = '{sas_token}';
        """

        create_external_db_sql = f"""
        CREATE EXTERNAL DATA SOURCE {external_database_name} WITH (
            LOCATION = f'https://{storage_account_name}.blob.core.windows.net/' \
            f'{container_name}',
            CREDENTIAL = {credential_name}
        );
        """

        self.run(create_master_key_sql)
        self.run(create_external_db_credential_sql)
        self.run(create_external_db_sql)

    def to_df(
        self,
        query: str,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        convert_bytes: bool = False,
        remove_special_characters: bool | None = None,
        columns_to_clean: list[str] | None = None,
    ) -> pd.DataFrame:
        """Execute a query and return the result as a pandas DataFrame.

        Args:
            query (str): The query to execute.
            con (pyodbc.Connection, optional): The connection to use to pull the data.
            if_empty (Literal["warn", "skip", "fail"], optional): What to do if the
                query returns no data. Defaults to None.
            convert_bytes (bool). A boolean value to trigger method
                df_converts_bytes_to_int. It is used to convert bytes data type into
                int, as pulling data with bytes can lead to malformed data in dataframe.
                Defaults to False.
            remove_special_characters (str, optional): Call a function that remove
                special characters like escape symbols. Defaults to None.
            columns_to_clean (List(str), optional): Select columns to clean, used with
                remove_special_characters. If None whole data frame will be processed.
                Defaults to None.
        """
        df = super().to_df(query=query, if_empty=if_empty)

        if convert_bytes:
            df = df_converts_bytes_to_int(df=df)

        if remove_special_characters:
            df = df_clean_column(df=df, columns_to_clean=columns_to_clean)

        return df



"""Flow for transforming data in DuckDB."""

from typing import Any

from prefect import flow

from viadot.orchestration.prefect.tasks import duckdb_query
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="transform--duckdb",
    description="Transform data in the DuckDB.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def duckdb_transform(
    query: str,
    duckdb_credentials_secret: str | None = None,
    # Specifying credentials in a dictionary is not recommended in the viadot flows,
    # but in this case credentials can include only database name.
    duckdb_credentials: dict[str, Any] | None = None,
    duckdb_config_key: str | None = None,
) -> None:
    """Transform data inside DuckDB database.

    Args:
        query (str, required): The query to execute on the DuckDB database.
        duckdb_credentials_secret (str, optional): The name of the secret storing
            the credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        duckdb_credentials (dict[str, Any], optional): Credentials to the database.
            Defaults to None.
        duckdb_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
    """
    duckdb_query(
        query=query,
        credentials=duckdb_credentials,
        config_key=duckdb_config_key,
        credentials_secret=duckdb_credentials_secret,
    )



"""Flow for transforming data inside the SQLServer."""

from prefect import flow

from viadot.orchestration.prefect.tasks import sql_server_query
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="transform--sql_server",
    description="Transform data inside the SQLServer.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def sql_server_transform(
    query: str,
    sql_server_credentials_secret: str | None = None,
    sql_server_config_key: str | None = None,
) -> None:
    """Run query inside the SQLServer.

    Args:
        query (str, required): The query to execute on the SQL Server database.
            If the qery doesn't start with "SELECT" returns an empty DataFrame.
        sql_server_credentials_secret (str, optional): The name of the secret storing
            the credentialsto the SQLServer. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        sql_server_config_key (str, optional): The key in the viadot config
            holding relevant credentials to the SQLServer. Defaults to None.
    """
    sql_server_query(
        query=query,
        config_key=sql_server_config_key,
        credentials_secret=sql_server_credentials_secret,
    )



"""Task to download data from SAP BW API into a Pandas DataFrame."""

from typing import Any

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_adls, sap_bw_to_df
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="SAP BW extraction to ADLS",
    description="Extract data from SAP BW and load it into Azure Data Lake Storage.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def sap_bw_to_adls(
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    mdx_query: str | None = None,
    mapping_dict: dict[str, Any] | None = None,
    adls_azure_key_vault_secret: str | None = None,
    adls_config_key: str | None = None,
    adls_path: str | None = None,
    adls_path_overwrite: bool = False,
) -> None:
    """Flow for downloading data from SAP BW API to Azure Data Lake.

    Args:
        config_key (Optional[str], optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret where credentials are stored. Defaults to None.
        mdx_query (str, optional): The MDX query to be passed to connection.
        mapping_dict (dict[str, Any], optional): Dictionary with original and new
            column names. Defaults to None.
        adls_azure_key_vault_secret (str, optional): The name of the Azure Key.
            Defaults to None.
        adls_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        adls_path (str, optional): Azure Data Lake destination folder/catalog path.
            Defaults to None.
        adls_path_overwrite (bool, optional): Whether to overwrite the file in ADLS.
            Defaults to False.
    """
    data_frame = sap_bw_to_df(
        config_key=config_key,
        azure_key_vault_secret=azure_key_vault_secret,
        mdx_query=mdx_query,
        mapping_dict=mapping_dict,
    )

    df_to_adls(
        df=data_frame,
        path=adls_path,
        credentials_secret=adls_azure_key_vault_secret,
        config_key=adls_config_key,
        overwrite=adls_path_overwrite,
    )



"""Extract data from OneStream SQL queries and load it to Redshift Spectrum."""

from typing import Any, Literal

from prefect import flow
from prefect.logging import get_run_logger

from viadot.orchestration.prefect.tasks import (
    create_batch_list_of_custom_subst_vars,
    df_to_redshift_spectrum,
    onestream_to_df,
)
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="extract--onestream_sql_query--redshift_spectrum",
    description="Extract data from OneStream SQL queries and load it to Redshift Spectrum.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def onestream_sql_query_data_to_redshift_spectrum(  # noqa: PLR0913
    base_url: str,
    application: str,
    sql_query: str,
    to_path: str,
    schema_name: str,
    table: str,
    custom_subst_vars: dict[str, list[Any]] | None = None,
    batch_by_subst_vars: bool = False,
    db_location: str = "Application",
    results_table_name: str = "Results",
    external_db: str = "",
    params: dict[str, str] | None = None,
    extension: str = ".parquet",
    if_exists: Literal["overwrite", "append"] = "overwrite",
    if_empty: Literal["warn", "skip", "fail"] = "fail",
    partition_cols: list[str] | None = None,
    index: bool = False,
    compression: Literal["snappy", "gzip", "zstd"] | None = None,
    sep: str = ",",
    aws_config_key: str | None = None,
    credentials_secret: str | None = None,
    onestream_credentials_secret: str | None = None,
    onestream_config_key: str | None = None,
) -> None:
    """Extract data from a OneStream SQL query data and load it into Redshift Spectrum.

    API documentation:
    https://documentation.onestream.com/1384528/Content/REST%20API/OneStream%20WebAPI%20Endpoints.html#:~:text=Data%20Provider%20GetAdoDataSetForSqlCommand%20endpoint

    This flow executes a SQL query in OneStream using provided parameters of the
    and uploads the results to AWS Redshift Spectrum.

    Flow can be executed with the custom substitution variables mapped to its values.
    These variable-value combination is used to filter data inside the data
    adapters table allowing to return only the datasets where rows have
    the value for a given column values set.The column - variable mapping is defined in
    the OneStream Data Adapter configuration.The filtering is done for each
    variable-values combination, hence the flow might generate a couple of API calls
    to get data for all combinations.

    When custom_subst_vars are provided and batch_by_subst_vars is True, the ingestion
    process will be split into batches. Each batch represents one combination of the
    custom_subst_vars - variable:[value]. When batch_by_subst_vars is False, all
    substitution variable combinations are processed together into a single data frame.
    Warning! Processing custom substitution vars without batching might lead to
    out-of-memory errors when each variable-value combination data set size is too big.


    Args:
        base_url (str): OneStream base server URL.
        application (str): OneStream application name.
        sql_query (str): SQL query to execute in OneStream.
        to_path (str): Path to an S3 folder where the table will be located.
        schema_name (str): AWS Glue catalog database name.
        table (str): AWS Glue catalog table name.
        custom_subst_vars (dict[str, list[Any]], optional): A dictionary mapping
            substitution variable names to lists of possible values.Substition variables
            used as substitution variables that are mapped to the column names in the
            Data Adapter configuration.They provides a kind of SQL WHERE clause for rows
            filtering where variable refers to a name of a set containings a list
            of possible values.
            For example: "prm_activity_region":["MK", "LG"]
                        refers to column UD3 that has custom variable mapping of
                        prm_activity_region.Data will be extracted only for rows where
                        the values (MK,LG) are present.
            Values can be of any type that can be converted to strings.Defaults to None.
        batch_by_subst_vars (bool): Whether to process data in batches based on
            substitution variable combinations. When True and custom_subst_vars is
            provided, each combination of substitution variables and its values will be
            processed as a separate batch, uploading individual parquet files in S3.
            When False,all substitution variable combinations are collected in a loop
            and are uploaded together in a single operation. Defaults to False.
        db_location (str): Database location path. Defaults to "Application".
        results_table_name (str): Results table name. Defaults to "Results".
        external_db (str): External database name. Defaults to "".
        params (dict[str, str], optional): Additional API parameters. Defaults to None.
        extension (str): Required file type. Accepted formats: 'csv', 'parquet'.
            Defaults to ".parquet".
        if_exists (Literal["overwrite", "append"], optional): Whether to 'overwrite'
            or 'append' to existing table. Defaults to "overwrite".
        if_empty (Literal["warn", "skip", "fail"], optional): What to do if the
            SQL query returns no data. Defaults to "fail".
        partition_cols (list[str], optional): Columns used to create partitions.
            Only applies when dataset=True. Defaults to None.
        index (bool): Write row names (index). Defaults to False.
        compression (Literal["snappy", "gzip", "zstd"], optional): Compression style
            (None, snappy, gzip, zstd). Defaults to None.
        sep (str): Field delimiter for the output file. Defaults to ','.
        aws_config_key (str, optional): Key in viadot config for AWS credentials.
            Defaults to None.
        credentials_secret (str, optional): Name of Prefect secret block with AWS
            credentials. Defaults to None.
        onestream_credentials_secret (str, optional): Name of secret storing OneStream
            credentials. Defaults to None.
        onestream_config_key (str, optional)): Key in viadot config for OneStream
            credentials.Defaults to None.
    """
    logger = get_run_logger()

    if batch_by_subst_vars and not custom_subst_vars:
        msg = (
            "Invalid parameter combination: batch_by_subst_vars=True requires "
            "custom_subst_vars to be provided. Either set batch_by_subst_vars=False "
            "or provide a custom_subst_vars dictionary."
        )
        logger.error(msg)
        raise ValueError(msg)

    logger.info(
        f"Starting OneStream SQL query execution from db_location: {db_location}."
    )

    if custom_subst_vars and batch_by_subst_vars:
        # Process data in batches - each substitution variable combination separately
        custom_subst_vars_batch_list = create_batch_list_of_custom_subst_vars(
            custom_subst_vars
        )
        logger.info(
            f"Processing {len(custom_subst_vars_batch_list)} batches based on substitution variable combinations."
        )

        for i, custom_subst_var in enumerate(custom_subst_vars_batch_list, 1):
            logger.info(
                f"Processing batch {i}/{len(custom_subst_vars_batch_list)}: {custom_subst_var}."
            )
            df = onestream_to_df(
                base_url=base_url,
                application=application,
                api="sql_query",
                sql_query=sql_query,
                custom_subst_vars=custom_subst_var,
                db_location=db_location,
                results_table_name=results_table_name,
                external_db=external_db,
                params=params,
                if_empty=if_empty,
                credentials_secret=onestream_credentials_secret,
                config_key=onestream_config_key,
            )
            df_to_redshift_spectrum(
                df=df,
                to_path=to_path,
                schema_name=schema_name,
                table=table,
                extension=extension,
                if_exists=if_exists,
                partition_cols=partition_cols,
                index=index,
                compression=compression,
                sep=sep,
                config_key=aws_config_key,
                credentials_secret=credentials_secret,
            )
            if_exists = "append"  # Changed to "append" to add batches to the table.
            logger.info(
                f"Batch {i}/{len(custom_subst_vars_batch_list)} completed successfully."
            )

        logger.info("All batches processed successfully.")
    else:
        # Process all data together - either no custom_subst_vars or batching disabled
        df = onestream_to_df(
            base_url=base_url,
            application=application,
            api="sql_query",
            sql_query=sql_query,
            custom_subst_vars=custom_subst_vars,
            db_location=db_location,
            results_table_name=results_table_name,
            external_db=external_db,
            params=params,
            if_empty=if_empty,
            credentials_secret=onestream_credentials_secret,
            config_key=onestream_config_key,
        )
        df_to_redshift_spectrum(
            df=df,
            to_path=to_path,
            schema_name=schema_name,
            table=table,
            extension=extension,
            if_exists=if_exists,
            partition_cols=partition_cols,
            index=index,
            compression=compression,
            sep=sep,
            config_key=aws_config_key,
            credentials_secret=credentials_secret,
        )
        logger.info("Data processing completed successfully.")



"""Exceptions to raise in case of issues while configuring or running viadot."""


class MissingSourceCredentialsError(Exception):
    """Raise when no source credentials were provided.

    Args:
        message (str, optional): A custom message to pass.
    """

    def __init__(self, message: str | None = None) -> None:
        """Override the default message."""
        default_message = "Either `credentials_secret`, `config_key`, or `credentials` has to be specified and not empty."
        message = message or default_message

        super().__init__(message)


class MissingPrefectBlockError(Exception):
    """Raise when a Prefect block is not found.

    Args:
        message (str, optional): A custom message to pass.
    """

    pass
